from __future__ import annotations

import enum
import math
import struct
from abc import ABC, abstractmethod
from typing import (
    AnyStr,
    Generic,
    Iterable,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)


# As a byte string to save space.
TAG_LENGTH = b"\x00\x01\x02\x04\x02\x04\x06\x08"
INT_SIZE = "BHIQ"


class ElementType(enum.IntEnum):
    SIGNED_INT = 0b00000
    UNSIGNED_INT = 0b00100
    BOOL = 0b01000
    FLOAT = 0b01010
    UTF8_STRING = 0b01100
    OCTET_STRING = 0b10000
    NULL = 0b10100
    STRUCTURE = 0b10101
    ARRAY = 0b10110
    LIST = 0b10111
    END_OF_CONTAINER = 0b11000


def decode_tag(control_octet, buffer, offset=0):
    tag_control = control_octet >> 5

    this_tag = None
    if tag_control == 0:  # Anonymous
        this_tag = None
    elif tag_control == 1:  # Context specific
        this_tag = buffer[offset]
    else:
        vendor_id = None
        profile_number = None
        if tag_control >= 6:  # Fully qualified
            vendor_id, profile_number = struct.unpack_from("<HH", buffer, offset)

        if tag_control in (0b010, 0b011):
            raise NotImplementedError("Common profile tag")

        if tag_control == 7:  # 4 octet tag number
            tag_number = struct.unpack_from("<I", buffer, offset + 4)[0]
        else:
            tag_number = struct.unpack_from("<H", buffer, offset + 4)[0]
        if vendor_id:
            this_tag = (vendor_id, profile_number, tag_number)
        else:
            this_tag = tag_number
    return this_tag, offset + TAG_LENGTH[tag_control]


def decode_element(control_octet, buffer, offset, depth):
    element_type = control_octet & 0x1F
    element_category = element_type >> 2
    if element_category == 0 or element_category == 1:  # ints
        member_class = NumberMember
    elif element_category == 2:  # Bool or float
        if element_type & 0x3 <= 1:
            member_class = BoolMember
        else:  # Float
            member_class = NumberMember
    elif element_type == 0b10100:  # Null
        member_class = None
    elif element_category == (ElementType.UTF8_STRING >> 2):
        member_class = UTF8StringMember
    elif element_category == (ElementType.OCTET_STRING >> 2):
        member_class = OctetStringMember
    elif element_type == ElementType.STRUCTURE:
        member_class = StructMember
    elif element_type == ElementType.ARRAY:
        member_class = ArrayMember
    elif element_type == ElementType.LIST:
        member_class = ListMember
    else:
        raise ValueError(f"Unknown element type {element_type:b}")

    if member_class is None:
        value = None
        offset = offset
    else:
        result = member_class.decode_member(control_octet, buffer, offset, depth)
        value, offset = result
    return value, offset


class Container:
    _max_length = None

    def __init__(self):
        self.values = {}

    @classmethod
    def max_length(cls):
        if cls._max_length is None:
            cls._max_length = sum(member.max_length for _, member in cls._members())
        return cls._max_length + 2

    @classmethod
    def _members(cls) -> Iterable[tuple[str, Member]]:
        for superclass in cls.__mro__:
            for field_name, descriptor in vars(superclass).items():
                if not field_name.startswith("_") and isinstance(descriptor, Member):
                    yield field_name, descriptor

    @classmethod
    def _members_by_tag(cls) -> dict[int, tuple[str, Member]]:
        if hasattr(cls, "_members_by_tag_cache"):
            return cls._members_by_tag_cache
        members = {}
        for field_name, descriptor in vars(cls).items():
            if not field_name.startswith("_") and isinstance(descriptor, Member):
                members[descriptor.tag] = (field_name, descriptor)
        cls._members_by_tag_cache = members
        return members

    def set_value(self, tag, value):
        self.values[tag] = value

    def delete_value(self, tag):
        if tag in self.values:
            del self.values[tag]


class Structure(Container):
    def __str__(self):
        members = []
        for field, descriptor_class in self._members():
            value = getattr(self, field)  # type: ignore  # self inference issues
            if value is None:
                if descriptor_class.optional:
                    continue
                value = "null"
            else:
                value = descriptor_class.print(value)
            if isinstance(descriptor_class, StructMember):
                value = value.replace("\n", "\n  ")
            members.append(f"{field} = {value}")
        return "{\n  " + ",\n  ".join(members) + "\n}"

    def encode(self) -> memoryview:
        buffer = bytearray(self.max_length())
        buffer[0] = ElementType.STRUCTURE
        end = self.encode_into(buffer, offset=1)
        return memoryview(buffer)[:end]

    def encode_into(self, buffer: bytearray, offset: int = 0) -> int:
        for _, descriptor_class in self._members():
            offset = descriptor_class.encode_into(self, buffer, offset)
        buffer[offset] = ElementType.END_OF_CONTAINER
        return offset + 1

    @classmethod
    def decode(cls, buffer: memoryview, offset=0) -> Structure:
        control_octet = buffer[offset]
        values, offset = cls.decode_member(control_octet, buffer, offset + 1)
        return values

    @classmethod
    def decode_member(
        cls, control_octet, buffer, offset=0, depth=0
    ) -> tuple[dict, int]:
        values = {}
        buffer = memoryview(buffer)
        while offset < len(buffer) and buffer[offset] != ElementType.END_OF_CONTAINER:
            control_octet = buffer[offset]
            this_tag, offset = decode_tag(control_octet, buffer, offset + 1)
            value, offset = decode_element(control_octet, buffer, offset, depth + 1)
            values[this_tag] = value

        if cls == Structure:
            return values, offset

        return cls.from_value(values), offset

    def construct_containers(self):
        tags = set(self.values.keys())
        for name, member_class in self._members():
            tag = member_class.tag
            if tag not in self.values:
                continue
            tags.remove(tag)
            self.values[tag] = member_class.from_value(self.values[tag])
        if tags:
            raise RuntimeError(f"Unknown tags {tags} in {type(self)}")

    @classmethod
    def from_value(cls, value):
        instance = cls()
        instance.values = value
        instance.construct_containers()
        return instance


_T = TypeVar("_T")
_NULLABLE = TypeVar("_NULLABLE", Literal[True], Literal[False])
_OPT = TypeVar("_OPT", Literal[True], Literal[False])


class Member(ABC, Generic[_T, _OPT, _NULLABLE]):
    max_value_length: int = 0

    def __init__(
        self, tag, *, optional: _OPT = False, nullable: _NULLABLE = False, default=None
    ) -> None:
        """
        :param optional: Indicates whether the value MAY be omitted from the encoding.
                         Can be used for deprecation.
        :param nullable: Indicates whether a TLV Null MAY be encoded in place of a value.
        """
        self.tag = tag
        self.optional = optional
        self.nullable = nullable
        self.tag_length = 0
        if isinstance(tag, int):
            self.tag_length = 1
            if tag >= 256:
                raise ValueError("Context specific tag too large")
        elif isinstance(tag, tuple):
            if tag[2] < 65536:
                self.tag_length = 6
            else:
                self.tag_length = 8
        self._max_length = None
        self._default = default
        self._name = None

    def __set_name__(self, owner, name):
        self._name = name

    @property
    def max_length(self):
        return 1 + self.tag_length + self.max_value_length

    @overload
    def __get__(
        self: Union[
            Member[_T, Literal[True], _NULLABLE], Member[_T, _OPT, Literal[True]]
        ],
        obj: Structure,
        objtype: Optional[Type[Structure]] = None,
    ) -> Optional[_T]: ...

    @overload
    def __get__(
        self: Member[_T, Literal[False], Literal[False]],
        obj: Structure,
        objtype: Optional[Type[Structure]] = None,
    ) -> _T: ...

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.tag
        if self.tag in obj.values:
            return obj.values[self.tag]
        return self._default

    @overload
    def __set__(
        self: Union[
            Member[_T, Literal[True], _NULLABLE], Member[_T, _OPT, Literal[True]]
        ],
        obj: Structure,
        value: Optional[_T],
    ) -> None: ...
    @overload
    def __set__(
        self: Member[_T, Literal[False], Literal[False]], obj: Structure, value: _T
    ) -> None: ...
    def __set__(self, obj, value):
        if value is None and not self.nullable:
            raise ValueError("Not nullable")
        obj.set_value(self.tag, value)

    def __delete__(self, obj):
        if not self.optional:
            raise ValueError("Not optional")
        obj.delete_value(self.tag)

    def encode(self, value):
        buffer = memoryview(bytearray(self.max_length))
        end = self._encode_value_into(value, buffer, 0, anonymous_ok=True)
        return buffer[:end]

    def encode_into(
        self,
        obj: Container,
        buffer: Union[bytearray, memoryview],
        offset: int,
        anonymous_ok=False,
    ) -> int:
        value = self.__get__(obj)  # type: ignore  # self inference issues
        return self._encode_value_into(value, buffer, offset, anonymous_ok)

    def _encode_value_into(
        self,
        value,
        buffer: Union[bytearray, memoryview],
        offset: int,
        anonymous_ok=False,
    ):
        element_type = ElementType.NULL
        if value is not None:
            element_type = self.encode_element_type(value)
        elif self.optional:
            # Value is None and the field is optional so skip it.
            return offset
        elif not self.nullable:
            raise ValueError(
                f"{self._name} ({type(self).__name__}) isn't set and not nullable or optional"
            )

        tag_control = 0
        if self.tag is not None:
            tag_control = 1
            if isinstance(self.tag, tuple):
                tag_control = 0b110
                if self.tag[2] >= 65536:
                    tag_control = 0b111
        elif not anonymous_ok:
            raise ValueError("Anonymous tag not allowed")

        buffer[offset] = tag_control << 5 | element_type
        offset += 1
        if self.tag is not None:
            if isinstance(self.tag, int):
                buffer[offset] = self.tag
                offset += 1
            else:
                vendor_id, profile_number, tag_number = self.tag
                struct.pack_into("<HH", buffer, offset, vendor_id, profile_number)
                offset += 4
                if tag_number >= 65536:
                    struct.pack_into("<I", buffer, offset, tag_number)
                    offset += 4
                else:
                    struct.pack_into("<H", buffer, offset, tag_number)
                    offset += 2
        if value is not None:
            new_offset = self.encode_value_into(  # type: ignore  # self inference issues
                value,
                buffer,
                offset,
            )
            return new_offset
        return offset

    def decode(self, buffer: memoryview, offset: int = 0) -> _T:
        "Return the decoded value at `offset` in `buffer`"
        return self.decode_member(buffer[offset], buffer, offset + 1)[0]

    @abstractmethod
    def decode_member(
        self, control_octet: int, buffer: memoryview, offset: int = 0
    ) -> (_T, int):
        "Return the decoded value at `offset` in `buffer`. `offset` is after the tag (but before any length)"
        ...

    @abstractmethod
    def encode_element_type(self, value: _T) -> int:
        "Return Element Type Field as defined in Appendix A in the spec"
        ...

    @overload
    @abstractmethod
    def encode_value_into(
        self: Union[
            Member[_T, Literal[True], _NULLABLE], Member[_T, _OPT, Literal[True]]
        ],
        value: Optional[_T],
        buffer: bytearray,
        offset: int,
    ) -> int: ...
    @overload
    @abstractmethod
    def encode_value_into(
        self: Member[_T, Literal[False], Literal[False]],
        value: _T,
        buffer: bytearray,
        offset: int,
    ) -> int: ...
    @abstractmethod
    def encode_value_into(
        self, value: Optional[_T], buffer: bytearray, offset: int
    ) -> int:
        "Encode `value` into `buffer` and return the new offset"
        ...

    @abstractmethod
    def print(self, value: _T) -> str:
        "Return string representation of `value`"
        ...

    def from_value(self, value):
        if value is None:
            if not self.nullable:
                raise ValueError("Member not nullable")
            return None
        return self._from_value(value)

    def _from_value(self, value):
        return value


# number type
_NT = TypeVar("_NT", float, int)


class NumberMember(Member[_NT, _OPT, _NULLABLE], Generic[_NT, _OPT, _NULLABLE]):
    def __init__(
        self,
        tag,
        _format: str,
        optional: _OPT = False,
        nullable: _NULLABLE = False,
        minimum: Optional[int] = None,
        maximum: Optional[int] = None,
        **kwargs,
    ):
        self.format = _format
        self.integer = _format[-1].upper() in INT_SIZE
        self.signed = self.format.islower()
        self.max_value_length = struct.calcsize(self.format)
        self._minimum = minimum
        self._maximum = maximum
        if self.integer:
            self._element_type = (
                ElementType.SIGNED_INT if self.signed else ElementType.UNSIGNED_INT
            )
            self._element_type |= int(math.log(self.max_value_length, 2))
        else:
            self._element_type = ElementType.FLOAT
            if self.max_value_length == 8:
                self._element_type |= 1
        super().__init__(tag, optional=optional, nullable=nullable, **kwargs)

    def __set__(self, obj, value):
        if value is not None and self.integer:
            octets = 2 ** INT_SIZE.index(self.format.upper()[-1])
            bits = 8 * octets
            max_size: int = (2 ** (bits - 1) if self.signed else 2**bits) - 1
            min_size: int = -max_size - 1 if self.signed else 0
            if not min_size <= value <= max_size:
                raise ValueError(
                    f"Out of bounds for {octets} octet {'' if self.signed else 'un'}signed int"
                )
        if self._minimum is not None and value < self._minimum:
            raise ValueError(f"Value is less than minimum of {self._minimum}")

        if self._maximum is not None and value > self._maximum:
            raise ValueError(f"Value is greater than maximum of {self._maximum}")

        super().__set__(obj, value)  # type: ignore  # self inference issues

    @staticmethod
    def decode_member(control_octet, buffer, offset=0, depth=0) -> tuple[_NT, int]:
        element_type = control_octet & 0x1F
        element_category = element_type >> 2
        if element_category == 0 or element_category == 1:
            length = 1 << (control_octet & 0x3)
            encoded_format = INT_SIZE[int(math.log(length, 2))]
            if element_category == 0:
                encoded_format = encoded_format.lower()
        else:
            length = 4 << (control_octet & 0x1)
            if length == 4:
                encoded_format = "<f"
            else:
                encoded_format = "<d"
        return (
            struct.unpack_from(encoded_format, buffer, offset=offset)[0],
            offset + struct.calcsize(encoded_format),
        )

    def print(self, value):
        unsigned = "" if self.signed else "U"
        return f"{value}{unsigned}"

    def encode_element_type(self, value):
        if self.integer:
            bit_length = value.bit_length()
            if self.signed:
                type = ElementType.SIGNED_INT
            else:
                type = ElementType.UNSIGNED_INT
            length = 0  # in power of two
            if bit_length <= 8:
                length = 0
            elif bit_length <= 16:
                length = 1
            elif bit_length <= 32:
                length = 2
            else:
                length = 3
            return type | length
        return self._element_type

    def encode_value_into(self, value, buffer, offset) -> int:
        if self.integer:
            bit_length = value.bit_length()
            format_string = None
            if bit_length <= 8:
                format_string = "<b" if self.signed else "<B"
                length = 1
            elif bit_length <= 16:
                format_string = "<h" if self.signed else "<H"
                length = 2
            elif bit_length <= 32:
                format_string = "<i" if self.signed else "<I"
                length = 4
            else:
                format_string = "<q" if self.signed else "<Q"
                length = 8
            struct.pack_into(format_string, buffer, offset, value)
            return offset + length
        # Float
        struct.pack_into(self.format, buffer, offset, value)
        return offset + self.max_value_length


class IntMember(NumberMember[int, _OPT, _NULLABLE]):
    def __init__(
        self,
        tag,
        *,
        signed: bool = True,
        octets: Literal[1, 2, 4, 8] = 1,
        optional: _OPT = False,
        nullable: _NULLABLE = False,
        **kwargs,
    ):
        """
        :param octets: Number of octests to use for encoding.
                       1, 2, 4, 8 are 8, 16, 32, and 64 bits respectively
        :param optional: Indicates whether the value MAY be omitted from the encoding.
                         Can be used for deprecation.
        :param nullable: Indicates whether a TLV Null MAY be encoded in place of a value.
        """
        # TODO 7.18.1 mentions other bit lengths (that are not a power of 2) than the TLV Appendix
        uformat = INT_SIZE[int(math.log2(octets))]
        # < = little-endian
        self.format = f"<{uformat.lower() if signed else uformat}"
        super().__init__(
            tag, _format=self.format, optional=optional, nullable=nullable, **kwargs
        )


class EnumMember(IntMember):
    def __init__(self, tag, enum_class, **kwargs):
        self.enum_class = enum_class
        super().__init__(tag, octets=2, signed=False, **kwargs)

    def __set__(self, obj, value):
        if not isinstance(value, self.enum_class):
            raise ValueError(f"Value must be a {self.enum_class}")
        super().__set__(obj, value.value)

    def __get__(self, obj, objtype=None) -> Optional[enum.Enum]:
        value = super().__get__(obj, objtype)
        if value is not None:
            return self.enum_class(value)
        return

    def print(self, value):
        return self.enum_class(value).name


class BitmapMember(EnumMember):
    def print(self, value):
        return repr(self.enum_class(value))


class FloatMember(NumberMember[float, _OPT, _NULLABLE]):
    def __init__(
        self,
        tag,
        *,
        octets: Literal[4, 8] = 4,
        optional: _OPT = False,
        nullable: _NULLABLE = False,
        **kwargs,
    ):
        """
        :param octets: Number of octests to use for encoding.
                       4, 8 are single and double precision floats respectively.
        :param optional: Indicates whether the value MAY be omitted from the encoding.
                         Can be used for deprecation.
        :param nullable: Indicates whether a TLV Null MAY be encoded in place of a value.
        """
        # < = little-endian
        self.format = f"<{'f' if octets == 4 else 'd'}"
        super().__init__(
            tag, _format=self.format, optional=optional, nullable=nullable, **kwargs
        )


class BoolMember(Member[bool, _OPT, _NULLABLE]):
    max_value_length = 0

    @staticmethod
    def decode_member(control_octet, buffer, offset=0, depth=0):
        return (control_octet & 1 == 1, offset)

    def print(self, value):
        if value:
            return "true"
        return "false"

    def encode_element_type(self, value):
        return ElementType.BOOL | (1 if value else 0)

    def encode_value_into(self, value, buffer, offset) -> int:
        return offset


class StringMember(Member[AnyStr, _OPT, _NULLABLE], Generic[AnyStr, _OPT, _NULLABLE]):
    _base_element_type: ElementType

    def __init__(
        self,
        tag,
        max_length: int,
        *,
        optional: _OPT = False,
        nullable: _NULLABLE = False,
        min_length: int = 0,
        **kwargs,
    ):
        self._element_type = self._base_element_type

        max_length_encoding = int(math.log(max_length, 256))
        max_length_format = INT_SIZE[max_length_encoding]
        self.length_length = struct.calcsize(max_length_format)
        self.min_length = min_length
        if max_length is None:
            raise ValueError("max_length is required")
        self._max_string_length = max_length if max_length is not None else 1280
        self.max_value_length = max_length + self.length_length
        super().__init__(tag, optional=optional, nullable=nullable, **kwargs)

    def print(self, value):
        return " ".join((f"{byte:02x}" for byte in value))

    def __set__(self, obj, value):
        if len(value) > self._max_string_length:
            raise ValueError(
                f"Value too long. {len(value)} > {self._max_string_length} bytes"
            )
        if len(value) < self.min_length:
            raise ValueError(f"Value too short. {len(value)} < {self.min_length} bytes")

        super().__set__(obj, value)  # type: ignore  # self inference issues

    def encode_element_type(self, value):
        # Log only works for 1+ so make 0 1 for length encoding.
        value_length = len(value)
        if value_length <= 0:
            value_length = 1
        length_encoding = int(math.log(value_length, 256))
        return self._element_type | length_encoding

    def encode_value_into(self, value, buffer: bytearray, offset: int) -> int:
        # Log only works for 1+ so make 0 1 for length encoding.
        value_length = len(value)
        if value_length <= 0:
            value_length = 1
        length_encoding = int(math.log(value_length, 256))
        length_format = INT_SIZE[length_encoding]
        length_length = struct.calcsize(length_format)
        struct.pack_into(length_format, buffer, offset, len(value))
        offset += length_length
        buffer[offset : offset + len(value)] = value
        return offset + len(value)

    @staticmethod
    def parse_length(control_octet, buffer, offset=0):
        element_type = control_octet & 0x1F
        power_of_two = element_type & 0x3
        length_length = 1 << power_of_two
        value_length = struct.unpack_from(INT_SIZE[power_of_two], buffer, offset)[0]
        return value_length, offset + length_length


class OctetStringMember(StringMember[bytes, _OPT, _NULLABLE]):
    _base_element_type: ElementType = ElementType.OCTET_STRING

    @staticmethod
    def decode_member(control_octet, buffer, offset=0, depth=0):
        length, offset = StringMember.parse_length(control_octet, buffer, offset)
        return (buffer[offset : offset + length].tobytes(), offset + length)


class UTF8StringMember(StringMember[str, _OPT, _NULLABLE]):
    _base_element_type = ElementType.UTF8_STRING

    @staticmethod
    def decode_member(control_octet, buffer, offset=0, depth=0):
        length, offset = StringMember.parse_length(control_octet, buffer, offset)
        return (
            buffer[offset : offset + length].tobytes().decode("utf-8"),
            offset + length,
        )

    def encode_value_into(self, value: str, buffer, offset) -> int:
        return super().encode_value_into(value.encode("utf-8"), buffer, offset)

    def print(self, value):
        return f'"{value}"'


_TLVStruct = TypeVar("_TLVStruct", bound=Structure)


class StructMember(Member[_TLVStruct, _OPT, _NULLABLE]):
    def __init__(
        self,
        tag,
        substruct_class: Type[_TLVStruct],
        *,
        optional: _OPT = False,
        nullable: _NULLABLE = False,
        **kwargs,
    ):
        self.substruct_class = substruct_class
        self.max_value_length = substruct_class.max_length() + 1
        super().__init__(tag, optional=optional, nullable=nullable, **kwargs)

    @staticmethod
    def decode_member(control_octet, buffer, offset=0, depth=0):
        value, offset = Structure.decode_member(control_octet, buffer, offset, depth)
        return value, offset + 1

    def print(self, value):
        return str(value)

    def encode_element_type(self, value):
        return ElementType.STRUCTURE

    def encode_value_into(self, value, buffer: bytearray, offset: int) -> int:
        offset = value.encode_into(buffer, offset)
        return offset

    def _from_value(self, value):
        return self.substruct_class.from_value(value)


class ArrayEncodingError(Exception):
    def __init__(self, index, offset):
        self.index = index
        """First index not encoded"""
        self.offset = offset


class ArrayMember(Member[_TLVStruct, _OPT, _NULLABLE]):
    def __init__(
        self,
        tag,
        substruct_class: Type[_TLVStruct, Member],
        *,
        max_length: Optional[int] = None,
        optional: _OPT = False,
        nullable: _NULLABLE = False,
        **kwargs,
    ):
        self.substruct_class = substruct_class
        self.max_value_length = 1280
        self.max_items = max_length
        super().__init__(tag, optional=optional, nullable=nullable, **kwargs)

    @staticmethod
    def decode_member(control_octet, buffer, offset=0, depth=0):
        entries = []
        while buffer[offset] != ElementType.END_OF_CONTAINER:
            control_octet = buffer[offset]
            value, offset = decode_element(control_octet, buffer, offset + 1, depth + 1)
            entries.append(value)
        return (entries, offset + 1)

    def print(self, value):
        s = ["["]
        items = []
        for v in value:
            items.append(str(v))
        s.append(", ".join(items))
        s.append("]")
        return "".join(s)

    def encode_element_type(self, value):
        return ElementType.ARRAY

    def encode_value_into(self, value, buffer: memoryview, offset: int) -> int:
        subbuffer = memoryview(buffer)[:-1]
        last_offset = offset
        for i, v in enumerate(value):
            if offset >= len(buffer) - 1:
                # If we run out of room, mark our end and raise an exception.
                buffer[offset] = ElementType.END_OF_CONTAINER
                raise ArrayEncodingError(i, offset + 1)
            try:
                if isinstance(self.substruct_class, Member):
                    buffer[offset] = self.substruct_class.encode_element_type(v)
                    offset = self.substruct_class.encode_value_into(
                        v, subbuffer, offset + 1
                    )
                else:
                    if isinstance(v, Structure):
                        buffer[offset] = ElementType.STRUCTURE
                    elif isinstance(v, List):
                        buffer[offset] = ElementType.LIST
                    offset = v.encode_into(subbuffer, offset + 1)
            except (ValueError, IndexError, struct.error):
                # If we run out of room, mark our end and raise an exception.
                buffer[offset] = ElementType.END_OF_CONTAINER
                raise ArrayEncodingError(i, last_offset + 1)
            last_offset = offset
        buffer[offset] = ElementType.END_OF_CONTAINER
        return offset + 1

    def _from_value(self, value):
        for i in range(len(value)):
            value[i] = self.substruct_class.from_value(value[i])
        return value


class List(Container):
    def __init__(self):
        self.items = []
        # items by tag. First occurence wins.
        self.values = {}

    def __iter__(self):
        return iter(self.items)

    def __str__(self):
        members = []
        member_by_tag = self._members_by_tag()
        for item in self.items:
            if isinstance(item, tuple):
                tag, value = item
                if tag in member_by_tag:
                    name, member = member_by_tag[tag]
                else:
                    name = tag
            else:
                name = None
                value = item

            if member:
                value = member.print(value)
                if not value:
                    continue
                if isinstance(member, StructMember):
                    value = value.replace("\n", "\n  ")
            if name:
                members.append(f"{name} = {value}")
            else:
                members.append(value)
        return "[[ " + ", ".join(members) + "]]"

    def encode(self) -> memoryview:
        buffer = bytearray(self.max_length())
        buffer[0] = ElementType.LIST
        end = self.encode_into(buffer, offset=1)
        return memoryview(buffer)[:end]

    def encode_into(self, buffer: bytearray, offset: int = 0) -> int:
        member_by_tag = self._members_by_tag()
        for item in self.items:
            if isinstance(item, tuple):
                tag, value = item
                if tag in member_by_tag:
                    name, member = member_by_tag[tag]
                else:
                    raise NotImplementedError("Unknown tag")
                offset = member.encode_into(self, buffer, offset, anonymous_ok=True)
            else:
                raise NotImplementedError("Anonymous list member")
        buffer[offset] = ElementType.END_OF_CONTAINER
        return offset + 1

    @classmethod
    def from_value(cls, value):
        instance = cls()
        instance.items = value
        instance.values = {}
        members_by_tag = cls._members_by_tag()
        for i, item in enumerate(value):
            if isinstance(item, tuple):
                tag, value = item
                if tag in members_by_tag:
                    value = members_by_tag[tag][1].from_value(value)
                    instance.items[i] = (tag, value)
                if tag in instance.values:
                    continue
                instance.values[tag] = value
        return instance

    def set_value(self, tag, value):
        if tag in self.values:
            i = self.items.index((tag, self.values[tag]))
            self.items[i] = (tag, value)
        else:
            self.items.append((tag, value))
        self.values[tag] = value

    def delete_value(self, tag):
        for item in self.items:
            if item[0] == tag:
                self.items.remove(item)
        del self.values[tag]

    def copy(self):
        new = type(self)()
        new.items.extend(self.items)
        new.values.update(self.values)
        return new


_TLVList = TypeVar("_TLVList", bound=List)


class ListMember(Member):
    def __init__(
        self,
        tag,
        substruct_class: Type[_TLVList],
        *,
        optional: _OPT = False,
        nullable: _NULLABLE = False,
        **kwargs,
    ):
        self.substruct_class = substruct_class
        self.max_value_length = substruct_class.max_length() + 1
        super().__init__(tag, optional=optional, nullable=nullable, **kwargs)

    @staticmethod
    def decode_member(control_octet, buffer, offset=0, depth=0):
        raw_list = []
        while buffer[offset] != ElementType.END_OF_CONTAINER:
            control_octet = buffer[offset]

            this_tag, offset = decode_tag(control_octet, buffer, offset + 1)
            value, offset = decode_element(control_octet, buffer, offset, depth + 1)

            if this_tag is None:
                raw_list.append(value)
            else:
                raw_list.append((this_tag, value))
        return raw_list, offset + 1

    def print(self, value):
        return str(value)

    def encode_element_type(self, value):
        return ElementType.LIST

    def encode_value_into(self, value, buffer: bytearray, offset: int) -> int:
        offset = value.encode_into(buffer, offset)
        return offset

    def _from_value(self, value):
        return self.substruct_class.from_value(value)


class AnythingMember(Member):
    """Stores a TLV encoded value."""

    def decode_member(self, control_octet, buffer, offset=0):
        return None

    def print(self, value):
        if isinstance(value, bytes):
            return value.hex(" ")
        if isinstance(value, memoryview):
            return value.hex(" ")
        return str(value)

    def encode_element_type(self, value):
        return value[0]

    def encode_value_into(self, value, buffer: bytearray, offset: int) -> int:
        value_length = len(value) - 1
        buffer[offset : offset + value_length] = memoryview(value)[1:]
        return offset + value_length



import json


class PersistentDictionary:
    """This acts like a dictionary and is persisted when values change."""

    def __init__(self, filename=None, root=None, state=None):
        self.filename = filename
        self.root = root
        self.dirty = False
        self.persisted = {}
        self._state: dict
        if self.root is None and filename:
            with open(self.filename, "r") as state_file:
                self._state = json.load(state_file)
        elif state is not None:
            self._state = state
        else:
            raise ValueError("Provide filename or (root and state)")

    def wrap(self, value):
        return value

    def __setitem__(self, key, value):
        self._state[key] = value
        if self.root:
            self.root.dirty = True
        else:
            self.dirty = True

    def __getitem__(self, key):
        value = self._state[key]
        if isinstance(value, dict):
            if key not in self.persisted:
                root = self.root if self.root else self
                self.persisted[key] = PersistentDictionary(root=root, state=value)
            return self.persisted[key]
        return value

    def __delitem__(self, key):
        del self._state[key]
        if self.root:
            self.root.dirty = True
        else:
            self.dirty = True

    def keys(self):
        return self._state.keys()

    def __iter__(self):
        return iter(self._state)

    def commit(self):
        if not self.dirty:
            return
        if self.root:
            self.root.commit()
            return
        with open(self.filename, "w") as state_file:
            json.dump(self._state, state_file, indent=1)
        self.dirty = False



import enum

from circuitmatter.data_model import (
    Cluster,
    NumberAttribute,
    BoolAttribute,
    UTF8StringAttribute,
    StructAttribute,
)
from circuitmatter import tlv


class ProductFinish(enum.IntEnum):
    OTHER = 0
    MATTE = 1
    SATIN = 2
    POLISHED = 3
    RUGGED = 4
    FABRIC = 5


class Color(enum.IntEnum):
    BLACK = 0
    NAVY = 1
    GREEN = 2
    TEAL = 3
    MAROON = 4
    PURPLE = 5
    OLIVE = 6
    GRAY = 7
    BLUE = 8
    LIME = 9
    AQUA = 10
    RED = 11
    FUCHSIA = 12
    YELLOW = 13
    WHITE = 14
    NICKEL = 15
    CHROME = 16
    BRASS = 17
    COPPER = 18
    SILVER = 19
    GOLD = 20


class BasicInformationCluster(Cluster):
    CLUSTER_ID = 0x0028

    class CapabilityMinima(tlv.Structure):
        CaseSessionsPerFabric = tlv.IntMember(
            0, signed=False, octets=2, minimum=3, default=3
        )
        SubscriptionsPerFabric = tlv.IntMember(
            1, signed=False, octets=2, minimum=3, default=3
        )

    class ProductAppearance(tlv.Structure):
        Finish = tlv.EnumMember(0, ProductFinish, default=ProductFinish.OTHER)
        PrimaryColor = tlv.EnumMember(1, Color, default=Color.BLACK)

    data_model_revision = NumberAttribute(0x00, signed=False, bits=16, default=16)
    vendor_name = UTF8StringAttribute(0x01, max_length=32, default="CircuitMatter")
    vendor_id = NumberAttribute(0x02, signed=False, bits=16)
    product_name = UTF8StringAttribute(0x03, max_length=32, default="Test Device")
    product_id = NumberAttribute(0x04, signed=False, bits=16)
    node_label = UTF8StringAttribute(0x05, max_length=32, default="")
    location = UTF8StringAttribute(0x06, max_length=2, default="XX")
    hardware_version = NumberAttribute(0x07, signed=False, bits=16, default=0)
    hardware_version_string = UTF8StringAttribute(
        0x08, min_length=1, max_length=64, default="Unknown"
    )
    software_version = NumberAttribute(0x09, signed=False, bits=32, default=0)
    software_version_string = UTF8StringAttribute(
        0x0A, min_length=1, max_length=64, default="Unknown"
    )
    manufacturing_date = UTF8StringAttribute(
        0x0B, min_length=8, max_length=16, default="Unknown"
    )
    part_number = UTF8StringAttribute(0x0C, max_length=32, default="")
    product_url = UTF8StringAttribute(
        0x0D, max_length=256, default="https://github.com/adafruit/circuitmatter"
    )
    product_label = UTF8StringAttribute(0x0E, max_length=64, default="")
    serial_number = UTF8StringAttribute(0x0F, max_length=32, default="")
    local_config_disabled = BoolAttribute(0x10, default=False)
    reachable = BoolAttribute(0x11, default=True)
    unique_id = UTF8StringAttribute(0x12, max_length=32, default="")
    capability_minima = StructAttribute(
        0x13, CapabilityMinima, default=CapabilityMinima()
    )
    product_appearance = StructAttribute(
        0x14, ProductAppearance, default=ProductAppearance()
    )
    specification_version = NumberAttribute(0x15, signed=False, bits=32, default=0)
    max_paths_per_invoke = NumberAttribute(0x16, signed=False, bits=16, default=1)



"""Simple LED on and off as a light."""

import circuitmatter as cm
from circuitmatter.device_types.lighting import on_off

import digitalio
import board


class LED(on_off.OnOffLight):
    def __init__(self, name, led):
        super().__init__(name)
        self._led = led
        self._led.direction = digitalio.Direction.OUTPUT

    def on(self):
        self._led.value = True

    def off(self):
        self._led.value = False


matter = cm.CircuitMatter()
led = LED("led1", digitalio.DigitalInOut(board.D19))
matter.add_device(led)
led = LED("led2", digitalio.DigitalInOut(board.D20))
matter.add_device(led)
while True:
    matter.process_packets()



import colorsys

from .color_temperature import ColorTemperatureLight


from circuitmatter.clusters.lighting import color_control


class ExtendedColorLight(ColorTemperatureLight):
    DEVICE_TYPE_ID = 0x010D
    REVISION = 4

    def __init__(self, name):
        super().__init__(name)

        self._color_control.feature_map |= (
            color_control.FeatureBitmap.HUE_SATURATION
            | color_control.FeatureBitmap.ENHANCED_HUE
            | color_control.FeatureBitmap.COLOR_LOOP
            | color_control.FeatureBitmap.XY
        )

        self._color_control.move_to_hue_and_saturation = (
            self._move_to_hue_and_saturation
        )

    def _move_to_hue_and_saturation(self, session, value):
        try:
            r, g, b = colorsys.hsv_to_rgb(value.Hue / 254, value.Saturation / 254, 1)
            self.color_rgb = int(r * 255) << 16 | int(g * 255) << 8 | int(b * 255)
        except Exception as e:
            print(f"Error setting color: {e}")
            return

        self._color_control.ColorMode = color_control.ColorMode.HUE_SATURATION
        self._color_control.CurrentHue = value.Hue
        self._color_control.CurrentSaturation = value.Saturation



from .dimmable import DimmableLight

from circuitmatter.clusters.lighting import color_control


class ColorTemperatureLight(DimmableLight):
    DEVICE_TYPE_ID = 0x010C
    REVISION = 4

    def __init__(self, name):
        super().__init__(name)

        self._color_control = color_control.ColorControl()
        self._color_control.feature_map |= color_control.FeatureBitmap.COLOR_TEMPERATURE

        self.servers.append(self._color_control)

    @property
    def color_rgb(self):
        raise NotImplementedError()

    @color_rgb.setter
    def color_rgb(self, value):
        raise NotImplementedError()



import time

from .message import Message, ExchangeFlags, ProtocolId
from .protocol import SecureProtocolOpcode
from .interaction_model import ChunkedMessage

# Section 4.12.8
MRP_MAX_TRANSMISSIONS = 5
"""The maximum number of transmission attempts for a given reliable message. The sender MAY choose this value as it sees fit."""

MRP_BACKOFF_BASE = 1.6
"""The base number for the exponential backoff equation."""

MRP_BACKOFF_JITTER = 0.25
"""The scaler for random jitter in the backoff equation."""

MRP_BACKOFF_MARGIN = 1.1
"""The scaler margin increase to backoff over the peer idle interval."""

MRP_BACKOFF_THRESHOLD = 1
"""The number of retransmissions before transitioning from linear to exponential backoff."""

MRP_STANDALONE_ACK_TIMEOUT_MS = 200
"""Amount of time to wait for an opportunity to piggyback an acknowledgement on an outbound message before falling back to sending a standalone acknowledgement."""


class Exchange:
    def __init__(self, session, initiator: bool, exchange_id: int, protocols):
        self.initiator = initiator
        self.exchange_id = exchange_id
        self.protocols = protocols
        self.session = session

        self.pending_acknowledgement = None
        """Message number that is waiting for an ack from us"""
        self.send_standalone_time = None

        self.next_retransmission_time = None
        """When to next resend the message that hasn't been acked"""
        self.pending_retransmission = None
        """Message that we've attempted to send but hasn't been acked"""
        self.pending_payloads = []

    def send(
        self,
        application_payload=None,
        protocol_id=None,
        protocol_opcode=None,
        reliable=True,
    ):
        if self.pending_retransmission is not None:
            raise RuntimeError("Cannot send a message while waiting for an ack.")
        message = Message()
        message.exchange_flags = ExchangeFlags(0)
        if self.initiator:
            message.exchange_flags |= ExchangeFlags.I
        if self.pending_acknowledgement is not None:
            message.exchange_flags |= ExchangeFlags.A
            self.send_standalone_time = None
            message.acknowledged_message_counter = self.pending_acknowledgement
            self.pending_acknowledgement = None
        if reliable:
            message.exchange_flags |= ExchangeFlags.R
            self.pending_retransmission = message
        message.source_node_id = self.session.local_node_id
        if protocol_id is None:
            protocol_id = application_payload.PROTOCOL_ID
        message.protocol_id = protocol_id
        if protocol_opcode is None:
            protocol_opcode = application_payload.PROTOCOL_OPCODE
        message.protocol_opcode = protocol_opcode
        message.exchange_id = self.exchange_id
        if isinstance(application_payload, ChunkedMessage):
            chunk = memoryview(bytearray(1280))[:1200]
            offset = application_payload.encode_into(chunk)
            if application_payload.MoreChunkedMessages:
                self.pending_payloads.insert(0, application_payload)
            message.application_payload = chunk[:offset]
        else:
            message.application_payload = application_payload
        self.session.send(message)

    def send_standalone(self):
        if self.pending_retransmission is not None:
            self.session.send(self.pending_retransmission)
            return
        self.send(
            protocol_id=ProtocolId.SECURE_CHANNEL,
            protocol_opcode=SecureProtocolOpcode.MRP_STANDALONE_ACK,
            reliable=False,
        )

    def queue(self, payload):
        self.pending_payloads.append(payload)

    def receive(self, message) -> bool:
        """Process the message and return if the packet should be dropped."""
        # Section 4.12.5.2.1
        if message.exchange_flags & ExchangeFlags.A:
            if message.acknowledged_message_counter is None:
                # Drop messages that are missing an acknowledgement counter.
                return True
            if (
                self.pending_retransmission is not None
                and message.acknowledged_message_counter
                != self.pending_retransmission.message_counter
            ):
                # Drop messages that have the wrong acknowledgement counter.
                return True
            self.pending_retransmission = None
            self.next_retransmission_time = None

        if message.protocol_id not in self.protocols:
            # Drop messages that don't match the protocols we're waiting for.
            return True

        # Section 4.12.5.2.2
        # Incoming packets that are marked Reliable.
        if message.exchange_flags & ExchangeFlags.R:
            if message.duplicate:
                # Send a standalone acknowledgement.
                self.send_standalone()
                return True
            if self.pending_acknowledgement is not None:
                # Send a standalone acknowledgement with the message counter we're about to overwrite.
                self.send_standalone()
            self.pending_acknowledgement = message.message_counter
            self.send_standalone_time = (
                time.monotonic() + MRP_STANDALONE_ACK_TIMEOUT_MS / 1000
            )

        if message.duplicate:
            return True
        return False



from . import crypto
from . import protocol
from . import tlv
from . import session

import hashlib
import struct

from cryptography.hazmat.primitives.ciphers.aead import AESCCM

from ecdsa.ellipticcurve import AbstractPoint, Point, PointJacobi
from ecdsa.curves import NIST256p


class PASEMessage(tlv.Structure):
    PROTOCOL_ID = protocol.ProtocolId.SECURE_CHANNEL


# pbkdfparamreq-struct => STRUCTURE [ tag-order ]
# {
# initiatorRandom
#  [1] : OCTET STRING [ length 32 ],
# initiatorSessionId
#  [2] : UNSIGNED INTEGER [ range 16-bits ],
# passcodeId
#  [3] : UNSIGNED INTEGER [ length 16-bits ],
# hasPBKDFParameters
#  [4] : BOOLEAN,
# initiatorSessionParams [5, optional] : session-parameter-struct
# }
class PBKDFParamRequest(PASEMessage):
    PROTOCOL_OPCODE = protocol.SecureProtocolOpcode.PBKDF_PARAM_REQUEST

    initiatorRandom = tlv.OctetStringMember(1, 32)
    initiatorSessionId = tlv.IntMember(2, signed=False, octets=2)
    passcodeId = tlv.IntMember(3, signed=False, octets=2)
    hasPBKDFParameters = tlv.BoolMember(4)
    initiatorSessionParams = tlv.StructMember(
        5, session.SessionParameterStruct, optional=True
    )


# Crypto_PBKDFParameterSet => STRUCTURE [ tag-order ]
# {
# iterations [1] : UNSIGNED INTEGER [ range 32-bits ],
# salt [2] : OCTET STRING [ length 16..32 ],
# }
class Crypto_PBKDFParameterSet(tlv.Structure):
    iterations = tlv.IntMember(1, signed=False, octets=4)
    salt = tlv.OctetStringMember(2, 32)


# pbkdfparamresp-struct => STRUCTURE [ tag-order ]
# {
# initiatorRandom
#  [1] : OCTET STRING [ length 32 ],
# responderRandom
#  [2] : OCTET STRING [ length 32 ],
# responderSessionId
#  [3] : UNSIGNED INTEGER [ range 16-bits ],
# pbkdf_parameters
#  [4] : Crypto_PBKDFParameterSet,
# responderSessionParams [5, optional] : session-parameter-struct
# }
class PBKDFParamResponse(PASEMessage):
    PROTOCOL_OPCODE = protocol.SecureProtocolOpcode.PBKDF_PARAM_RESPONSE
    initiatorRandom = tlv.OctetStringMember(1, 32)
    responderRandom = tlv.OctetStringMember(2, 32)
    responderSessionId = tlv.IntMember(3, signed=False, octets=2)
    pbkdf_parameters = tlv.StructMember(4, Crypto_PBKDFParameterSet)
    responderSessionParams = tlv.StructMember(
        5, session.SessionParameterStruct, optional=True
    )


class PAKE1(PASEMessage):
    PROTOCOL_OPCODE = protocol.SecureProtocolOpcode.PASE_PAKE1
    pA = tlv.OctetStringMember(1, crypto.PUBLIC_KEY_SIZE_BYTES)


class PAKE2(PASEMessage):
    PROTOCOL_OPCODE = protocol.SecureProtocolOpcode.PASE_PAKE2
    pB = tlv.OctetStringMember(1, crypto.PUBLIC_KEY_SIZE_BYTES)
    cB = tlv.OctetStringMember(2, crypto.HASH_LEN_BYTES)


class PAKE3(PASEMessage):
    PROTOCOL_OPCODE = protocol.SecureProtocolOpcode.PASE_PAKE3
    cA = tlv.OctetStringMember(1, crypto.HASH_LEN_BYTES)


M = PointJacobi.from_bytes(
    NIST256p.curve,
    b"\x02\x88\x6e\x2f\x97\xac\xe4\x6e\x55\xba\x9d\xd7\x24\x25\x79\xf2\x99\x3b\x64\xe1\x6e\xf3\xdc\xab\x95\xaf\xd4\x97\x33\x3d\x8f\xa1\x2f",
)
N = PointJacobi.from_bytes(
    NIST256p.curve,
    b"\x03\xd8\xbb\xd6\xc6\x39\xc6\x29\x37\xb0\x4d\x99\x7f\x38\xc3\x77\x07\x19\xc6\x29\xd7\x01\x4d\x49\xa2\x4b\x4f\x98\xba\xa1\x29\x2b\x49",
)
crypto.W_SIZE_BYTES = crypto.GROUP_SIZE_BYTES + 8


# in the spake2p math P is NIST256p.generator
# in the spake2p math p is NIST256p.order
def _pbkdf2(passcode, salt, iterations):
    ws = hashlib.pbkdf2_hmac(
        "sha256", struct.pack("<I", passcode), salt, iterations, crypto.W_SIZE_BYTES * 2
    )
    w0 = int.from_bytes(ws[: crypto.W_SIZE_BYTES], byteorder="big") % NIST256p.order
    w1 = int.from_bytes(ws[crypto.W_SIZE_BYTES :], byteorder="big") % NIST256p.order
    return w0, w1


def initiator_values(passcode, salt, iterations) -> tuple[bytes, bytes]:
    w0, w1 = _pbkdf2(passcode, salt, iterations)
    return w0.to_bytes(NIST256p.baselen, byteorder="big"), w1.to_bytes(
        NIST256p.baselen, byteorder="big"
    )


# w0 and w1 are big-endian encoded
def Crypto_pA(w0, w1) -> bytes:
    return b""


def Crypto_pB(random_source, w0: int, L: Point) -> tuple[int, AbstractPoint]:
    y = random_source.randbelow(NIST256p.order)
    Y = y * NIST256p.generator + w0 * N
    return y, Y


def Crypto_Transcript(context, pA, pB, Z, V, w0) -> bytes:
    elements = [
        context,
        b"",
        b"",
        M.to_bytes("uncompressed"),
        N.to_bytes("uncompressed"),
        pA,
        pB,
        Z,
        V,
        w0,
    ]
    total_length = 0
    for e in elements:
        total_length += len(e) + 8
    tt = bytearray(total_length)
    offset = 0
    for e in elements:
        struct.pack_into("<Q", tt, offset, len(e))
        offset += 8

        tt[offset : offset + len(e)] = e
        offset += len(e)
    return tt


def KDF(salt, key, info):
    # Section 3.10 defines the mapping from KDF to Crypto_KDF but it is wrong!
    # The arg order is correct above.
    return crypto.KDF(key, salt, info, crypto.HASH_LEN_BITS)


def Crypto_P2(tt, pA, pB) -> tuple[bytes, bytes, bytes]:
    KaKe = crypto.Hash(tt)
    Ka = KaKe[: crypto.HASH_LEN_BYTES // 2]
    Ke = KaKe[crypto.HASH_LEN_BYTES // 2 :]
    # https://github.com/project-chip/connectedhomeip/blob/c88d5cf83cd3e3323ac196630acc34f196a2f405/src/crypto/CHIPCryptoPAL.cpp#L458-L468
    KcAKcB = KDF(None, Ka, b"ConfirmationKeys")
    KcA = KcAKcB[: crypto.HASH_LEN_BYTES // 2]
    KcB = KcAKcB[crypto.HASH_LEN_BYTES // 2 :]
    cA = crypto.HMAC(KcA, pB)
    cB = crypto.HMAC(KcB, pA)
    return (cA, cB, Ke)


def compute_session_keys(Ke, secure_session_context):
    keys = crypto.KDF(
        Ke,
        b"",
        b"SessionKeys",
        3 * crypto.SYMMETRIC_KEY_LENGTH_BITS,
    )
    secure_session_context.i2r_key = keys[: crypto.SYMMETRIC_KEY_LENGTH_BYTES]
    secure_session_context.i2r = AESCCM(
        secure_session_context.i2r_key,
        tag_length=crypto.AEAD_MIC_LENGTH_BYTES,
    )
    secure_session_context.r2i_key = keys[
        crypto.SYMMETRIC_KEY_LENGTH_BYTES : 2 * crypto.SYMMETRIC_KEY_LENGTH_BYTES
    ]
    secure_session_context.r2i = AESCCM(
        secure_session_context.r2i_key,
        tag_length=crypto.AEAD_MIC_LENGTH_BYTES,
    )
    secure_session_context.attestation_challenge = keys[
        2 * crypto.SYMMETRIC_KEY_LENGTH_BYTES : 3 * crypto.SYMMETRIC_KEY_LENGTH_BYTES
    ]


def compute_verification(random_source, pake1, pake2, context, verifier):
    w0 = memoryview(verifier)[: crypto.GROUP_SIZE_BYTES]
    L = memoryview(verifier)[crypto.GROUP_SIZE_BYTES :]
    L = Point.from_bytes(NIST256p.curve, L)
    w0 = int.from_bytes(w0, byteorder="big")
    y, Y = Crypto_pB(random_source, w0, L)
    # pB is Y encoded uncompressed
    # pA is X encoded uncompressed
    pake2.pB = Y.to_bytes("uncompressed")
    h = NIST256p.curve.cofactor()
    # Use negation because the class doesn't support subtraction. 🤦
    X = Point.from_bytes(NIST256p.curve, pake1.pA)
    Z = h * y * (X + (-(w0 * M)))
    # Z is wrong. V is right
    V = h * y * L
    tt = Crypto_Transcript(
        context,
        pake1.pA,
        pake2.pB,
        Z.to_bytes("uncompressed"),
        V.to_bytes("uncompressed"),
        w0.to_bytes(NIST256p.baselen, byteorder="big"),
    )
    cA, cB, Ke = Crypto_P2(tt, pake1.pA, pake2.pB)
    pake2.cB = cB
    return cA, Ke


def _write_bits(buf, offset, bits, value) -> int:
    while bits > 0:
        bits_remaining = 8 - offset % 8
        write_bits = min(bits, bits_remaining)
        mask = (1 << write_bits) - 1
        buf[offset // 8] |= (value & mask) << (offset % 8)
        offset += write_bits
        bits -= write_bits
        value >>= write_bits
    return offset


def _base38_encode(buf) -> str:
    alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ-."
    encoded = []
    for i in range(0, len(buf), 3):
        value = 0
        remaining = min(3, len(buf) - i)
        for j in range(remaining):
            value |= buf[i + j] << (j * 8)
        outputs = 5
        if remaining == 2:
            outputs = 4
        elif remaining == 1:
            outputs = 2
        for j in range(outputs):
            encoded.append(alphabet[value % 38])
            value //= 38
    return "".join(encoded)


def compute_qr_code(vendor_id, product_id, discriminator, passcode) -> str:
    total_bits = 3 + 16 * 2 + 2 + 8 + 12 + 27 + 4
    total_bytes = total_bits // 8
    buf = bytearray(total_bytes)

    discovery = 1 << 2  # On network already

    offset = 0
    offset = _write_bits(buf, offset, 3, 0)
    offset = _write_bits(buf, offset, 16, vendor_id)
    offset = _write_bits(buf, offset, 16, product_id)
    offset = _write_bits(buf, offset, 2, 0)
    offset = _write_bits(buf, offset, 8, discovery)
    offset = _write_bits(buf, offset, 12, discriminator)
    offset = _write_bits(buf, offset, 27, passcode)

    return _base38_encode(buf)


def show_qr_code(vendor_id, product_id, discriminator, passcode):
    encoded = compute_qr_code(vendor_id, product_id, discriminator, passcode)
    import qrcode

    qr = qrcode.QRCode(
        version=1,
        error_correction=qrcode.constants.ERROR_CORRECT_L,
        box_size=10,
        border=4,
    )
    qr.add_data("MT:")
    qr.add_data(encoded)
    print("QR code data: MT:" + encoded)
    qr.print_ascii()



from circuitmatter import data_model
from circuitmatter import tlv


class BindingCluster(data_model.Cluster):
    CLUSTER_ID = 0x001E
    cluster_revision = 1

    class TargetStruct(tlv.Structure):
        Node = data_model.NodeId(1, optional=True)
        Group = data_model.GroupId(2, optional=True)
        Endpoint = data_model.EndpointNumber(3, optional=True)
        Cluster = data_model.ClusterId(4, optional=True)

    Binding = data_model.ListAttribute(
        0x0000, TargetStruct, default=[], N_nonvolatile=True
    )



import binascii


class ReplaySocket:
    def __init__(self, replay_data):
        self.replay_data = replay_data

    def bind(self, address):
        print("bind to", address)

    def setblocking(self, value):
        print("setblocking", value)

    def recvfrom_into(self, buffer, nbytes=None):
        if nbytes is None:
            nbytes = len(buffer)
        direction = "send"
        while direction == "send":
            direction, _, address, data_b64 = self.replay_data.pop(0)
        decoded = binascii.a2b_base64(data_b64)
        if len(decoded) > nbytes:
            raise RuntimeError("Next replay packet is larger than buffer to read into")
        buffer[: len(decoded)] = decoded
        return len(decoded), address

    def sendto(self, data, address):
        if address is None:
            raise ValueError("Address must be set")
        # direction, _, address, data_b64 = self.replay_data.pop(0)
        # if direction == "send":
        #     decoded = binascii.a2b_base64(data_b64)
        # for i, b in enumerate(data):
        #     if b != decoded[i]:
        #         # print("sent", data.hex(" "))
        #         # print("old ", decoded.hex(" "))
        #         # print(i, hex(b), hex(decoded[i]))
        #         print("Next replay packet does not match sent data")
        return len(data)


class ReplayRandom:
    def __init__(self, replay_data):
        self.replay_data = replay_data

    def urandom(self, nbytes):
        direction = None
        while direction != "urandom":
            direction, _, recorded_nbytes, data_b64 = self.replay_data.pop(0)
            if recorded_nbytes != nbytes:
                raise RuntimeError("Next replay random data is not the expected length")
        decoded = binascii.a2b_base64(data_b64)
        return decoded

    def randbelow(self, n):
        direction = None
        while direction != "randbelow":
            direction, _, recorded_n, value = self.replay_data.pop(0)
            if recorded_n != n:
                raise RuntimeError("Next replay randbelow is not the expected length")
        return value


class ReplaySocketPool:
    AF_INET6 = 0
    SOCK_DGRAM = 1

    def __init__(self, replay_lines):
        self.replay_data = replay_lines
        self._socket_created = False

    def socket(self, *args, **kwargs):
        if self._socket_created:
            raise RuntimeError("Only one socket can be created")
        self._socket_created = True
        return ReplaySocket(self.replay_data)



from circuitmatter import interaction_model
from circuitmatter import tlv


class Message(interaction_model.ChunkedMessage):
    Array = tlv.ArrayMember(0, tlv.IntMember(None, signed=False, octets=2))
    MoreChunkedMessages = tlv.BoolMember(1, optional=True)
    After = tlv.BoolMember(2)  # Only one bool can fit after.


def fill_array(length):
    t = Message()
    t.Array = list(range(length))
    t.After = False
    return t


def test_only_one():
    buf = bytearray(20)
    t = fill_array(4)
    end = t.encode_into(buf)
    print(buf[:end].hex("x"))
    assert (
        buf[:end]
        == b"\x15\x36\x00\x04\x00\x04\x01\x04\x02\x04\x03\x18\x28\x02\x24\xff\x0b\x18"
    )


def test_two_chunks():
    buf = bytearray(16)
    t = fill_array(4)
    end = t.encode_into(buf)
    print(buf[:end].hex("x"))
    assert (
        buf[:end] == b"\x15\x36\x00\x04\x00\x04\x01\x18\x29\x01\x28\x02\x24\xff\x0b\x18"
    )

    buf = bytearray(16)
    end = t.encode_into(buf)
    print(buf[:end].hex("x"))
    assert buf[:end] == b"\x15\x36\x00\x04\x02\x04\x03\x18\x28\x02\x24\xff\x0b\x18"


def test_two_chunks_odd():
    buf = bytearray(17)
    t = fill_array(4)
    end = t.encode_into(buf)
    print(buf[:end].hex("x"))
    assert (
        buf[:end] == b"\x15\x36\x00\x04\x00\x04\x01\x18\x29\x01\x28\x02\x24\xff\x0b\x18"
    )

    buf = bytearray(17)
    end = t.encode_into(buf)
    print(buf[:end].hex("x"))
    assert buf[:end] == b"\x15\x36\x00\x04\x02\x04\x03\x18\x28\x02\x24\xff\x0b\x18"


def test_three_chunks():
    pass



import binascii
import ecdsa
from ecdsa import der
import hashlib
import struct
import time

from circuitmatter import crypto, interaction_model, tlv
from circuitmatter.clusters.device_management.basic_information import (
    BasicInformationCluster,
)
from circuitmatter.clusters.device_management.general_diagnostics import (
    GeneralDiagnosticsCluster,
)
from circuitmatter.clusters.device_management.general_commissioning import (
    GeneralCommissioningCluster,
    CommissioningErrorEnum,
)
from circuitmatter.clusters.device_management.group_key_management import (
    GroupKeyManagementCluster,
    GroupKeySetSecurityPolicyEnum,
    GroupKeySetStruct,
)
from circuitmatter.clusters.device_management.network_commissioning import (
    NetworkCommissioningCluster,
)
from circuitmatter.clusters.device_management.node_operational_credentials import (
    CertificateChainTypeEnum,
    NodeOperationalCredentialsCluster,
    NodeOperationalCertStatusEnum,
)
from circuitmatter.clusters.system_model import user_label
from circuitmatter.clusters.system_model.access_control import AccessControlCluster

from .. import simple_device

PAI_CERT_DER = b"\x30\x82\x01\xaa\x30\x82\x01\x50\xa0\x03\x02\x01\x02\x02\x08\x5e\xf5\xaf\xd0\x13\x60\xf5\xd4\x30\x0a\x06\x08\x2a\x86\x48\xce\x3d\x04\x03\x02\x30\x1a\x31\x18\x30\x16\x06\x03\x55\x04\x03\x0c\x0f\x4d\x61\x74\x74\x65\x72\x20\x54\x65\x73\x74\x20\x50\x41\x41\x30\x20\x17\x0d\x32\x34\x31\x30\x31\x37\x30\x30\x30\x30\x30\x30\x5a\x18\x0f\x39\x39\x39\x39\x31\x32\x33\x31\x32\x33\x35\x39\x35\x39\x5a\x30\x32\x31\x1a\x30\x18\x06\x03\x55\x04\x03\x0c\x11\x43\x69\x72\x63\x75\x69\x74\x4d\x61\x74\x74\x65\x72\x20\x50\x41\x49\x31\x14\x30\x12\x06\x0a\x2b\x06\x01\x04\x01\x82\xa2\x7c\x02\x01\x0c\x04\x46\x46\x46\x34\x30\x59\x30\x13\x06\x07\x2a\x86\x48\xce\x3d\x02\x01\x06\x08\x2a\x86\x48\xce\x3d\x03\x01\x07\x03\x42\x00\x04\x37\x5d\x2b\xc8\xc6\x15\x27\x5b\xfd\x84\x8b\x52\xfe\x21\x96\xe2\xa1\x4e\xf3\xcc\x91\xae\xf0\x5d\xff\x85\x1c\xbc\x19\xb1\xa9\x35\x45\x8c\xfe\x04\xaa\x42\x4e\x01\x6d\xe3\xd6\x74\xdc\x5b\x73\x29\xbd\x77\x57\xfd\xdb\x32\x38\xd6\x26\x73\x62\x9b\x3c\x79\x08\x45\xa3\x66\x30\x64\x30\x12\x06\x03\x55\x1d\x13\x01\x01\xff\x04\x08\x30\x06\x01\x01\xff\x02\x01\x00\x30\x0e\x06\x03\x55\x1d\x0f\x01\x01\xff\x04\x04\x03\x02\x01\x06\x30\x1d\x06\x03\x55\x1d\x0e\x04\x16\x04\x14\x07\xf8\x38\x0a\x5f\x01\x36\xfc\xe2\x36\xbd\x45\xf2\x88\xff\x22\xdc\xa6\xf4\xa7\x30\x1f\x06\x03\x55\x1d\x23\x04\x18\x30\x16\x80\x14\x78\x5c\xe7\x05\xb8\x6b\x8f\x4e\x6f\xc7\x93\xaa\x60\xcb\x43\xea\x69\x68\x82\xd5\x30\x0a\x06\x08\x2a\x86\x48\xce\x3d\x04\x03\x02\x03\x48\x00\x30\x45\x02\x21\x00\x9c\x5f\x59\x83\x5c\xc6\x51\xc2\x5c\x79\x01\x33\x25\x22\x3d\x25\x6b\xe7\x43\x98\xbc\x03\x83\x89\xf4\x55\x6d\xf7\xf7\x4a\x8a\x34\x02\x20\x5c\x14\x17\x4c\xc3\x23\x07\xff\x42\x1c\x4f\x8b\x0b\x63\xb9\x62\x52\x58\xa2\x96\xe0\x31\xfd\xce\x51\xa2\x7a\x08\x49\x2b\xc0\x38"


class _GeneralCommissioningCluster(GeneralCommissioningCluster):
    def __init__(self):
        super().__init__()
        basic_commissioning_info = GeneralCommissioningCluster.BasicCommissioningInfo()
        basic_commissioning_info.FailSafeExpiryLengthSeconds = 10
        basic_commissioning_info.MaxCumulativeFailsafeSeconds = 900
        self.basic_commissioning_info = basic_commissioning_info

    def arm_fail_safe(
        self, session, args: GeneralCommissioningCluster.ArmFailSafe
    ) -> GeneralCommissioningCluster.ArmFailSafeResponse:
        self.breadcrumb = args.Breadcrumb
        response = GeneralCommissioningCluster.ArmFailSafeResponse()
        response.ErrorCode = CommissioningErrorEnum.OK
        return response

    def set_regulatory_config(
        self, session, args: GeneralCommissioningCluster.SetRegulatoryConfig
    ) -> GeneralCommissioningCluster.SetRegulatoryConfigResponse:
        self.breadcrumb = args.Breadcrumb
        response = GeneralCommissioningCluster.SetRegulatoryConfigResponse()
        response.ErrorCode = CommissioningErrorEnum.OK
        return response

    def commissioning_complete(
        self, session
    ) -> GeneralCommissioningCluster.CommissioningCompleteResponse:
        response = GeneralCommissioningCluster.CommissioningCompleteResponse()
        response.ErrorCode = CommissioningErrorEnum.OK
        print("Commissioning done!")
        return response


class AttestationElements(tlv.Structure):
    certification_declaration = tlv.OctetStringMember(0x01, max_length=400)
    attestation_nonce = tlv.OctetStringMember(0x02, max_length=32)
    timestamp = tlv.IntMember(0x03, signed=False, octets=4)
    firmware_information = tlv.OctetStringMember(0x04, max_length=16, optional=True)
    """Used for secure boot. We don't support it."""


class NOCSRElements(tlv.Structure):
    csr = tlv.OctetStringMember(0x01, max_length=1024)
    CSRNonce = tlv.OctetStringMember(0x02, max_length=32)
    # Skip vendor reserved


def encode_set(*encoded_pieces):
    total_len = sum([len(p) for p in encoded_pieces])
    return b"\x31" + der.encode_length(total_len) + b"".join(encoded_pieces)


def encode_utf8_string(s):
    encoded = s.encode("utf-8")
    return b"\x0c" + der.encode_length(len(encoded)) + encoded


class _NodeOperationalCredentialsCluster(NodeOperationalCredentialsCluster):
    def __init__(self, group_key_manager, random_source, mdns_server, port):
        super().__init__()

        self.group_key_manager = group_key_manager

        self.dac_key = None

        self.new_key_for_update = False
        self.pending_root_cert = None
        self.pending_signing_key = None

        self.supported_fabrics = 10

        self.root_certs = []
        self.compressed_fabric_ids = []
        self.noc_keys = []
        self.encoded_noc_keys = []

        self.mdns_server = mdns_server
        self.port = port
        self.random = random_source

    def restore(self, nonvolatile):
        super().restore(nonvolatile)

        self.dac_key = ecdsa.keys.SigningKey.from_der(
            binascii.a2b_base64(nonvolatile["dac_key"]), hashfunc=hashlib.sha256
        )

        if "pk" not in nonvolatile:
            return

        self.root_certs = []
        self.compressed_fabric_ids = []
        self.noc_keys = []
        self.encoded_noc_keys = nonvolatile["pk"]

        for i, encoded_root_cert in enumerate(self.trusted_root_certificates):
            root_cert = crypto.MatterCertificate.decode(encoded_root_cert)

            self.root_certs.append(root_cert)
            fabric = self.fabrics[i]
            fabric_id = struct.pack(">Q", fabric.FabricID)
            compressed_fabric_id = crypto.KDF(
                root_cert.ec_pub_key[1:], fabric_id, b"CompressedFabric", 64
            )
            self.compressed_fabric_ids.append(compressed_fabric_id)
            signing_key = ecdsa.keys.SigningKey.from_string(
                binascii.a2b_base64(self.encoded_noc_keys[i]),
                curve=ecdsa.NIST256p,
                hashfunc=hashlib.sha256,
            )
            self.noc_keys.append(signing_key)

            node_id = struct.pack(">Q", fabric.NodeID).hex().upper()
            compressed_fabric_id = compressed_fabric_id.hex().upper()
            instance_name = f"{compressed_fabric_id}-{node_id}"
            self.mdns_server.advertise_service(
                "_matter",
                "_tcp",
                self.port,
                instance_name=instance_name,
                subtypes=[
                    f"_I{compressed_fabric_id}._sub._matter._tcp",
                ],
            )

    def certificate_chain_request(
        self,
        session,
        args: NodeOperationalCredentialsCluster.CertificateChainRequest,
    ) -> NodeOperationalCredentialsCluster.CertificateChainResponse:
        response = NodeOperationalCredentialsCluster.CertificateChainResponse()
        if args.CertificateType == CertificateChainTypeEnum.PAI:
            response.Certificate = PAI_CERT_DER
        elif args.CertificateType == CertificateChainTypeEnum.DAC:
            response.Certificate = binascii.a2b_base64(self._nonvolatile["dac_cert"])
        return response

    def attestation_request(
        self,
        session,
        args: NodeOperationalCredentialsCluster.AttestationRequest,
    ) -> NodeOperationalCredentialsCluster.AttestationResponse:
        elements = AttestationElements()
        elements.certification_declaration = binascii.a2b_base64(
            self._nonvolatile["cd"]
        )
        elements.attestation_nonce = args.AttestationNonce
        elements.timestamp = int(time.time())
        elements = elements.encode()
        attestation_tbs = elements.tobytes() + session.attestation_challenge
        response = NodeOperationalCredentialsCluster.AttestationResponse()
        response.AttestationElements = elements
        response.AttestationSignature = self.dac_key.sign_deterministic(
            attestation_tbs,
            hashfunc=hashlib.sha256,
            sigencode=ecdsa.util.sigencode_string,
        )
        return response

    def csr_request(
        self, session, args: NodeOperationalCredentialsCluster.CSRRequest
    ) -> NodeOperationalCredentialsCluster.CSRResponse:
        # Section 6.4.6.1
        # CSR stands for Certificate Signing Request. A NOCSR is a Node Operational Certificate Signing Request

        self.new_key_for_update = args.IsForUpdateNOC
        self.pending_signing_key = ecdsa.keys.SigningKey.generate(
            curve=ecdsa.NIST256p, hashfunc=hashlib.sha256, entropy=self.random.urandom
        )

        # DER encode the request
        # https://www.rfc-editor.org/rfc/rfc2986 Section 4.2
        certification_request = []

        certification_request_info = []

        # Version
        certification_request_info.append(der.encode_integer(0))

        # subject
        attribute_type = der.encode_oid(2, 5, 4, 10)
        value = encode_utf8_string("CSA")

        subject = der.encode_sequence(
            encode_set(der.encode_sequence(attribute_type, value))
        )
        certification_request_info.append(subject)

        # Subject Public Key Info
        algorithm = der.encode_sequence(
            der.encode_oid(1, 2, 840, 10045, 2, 1),
            der.encode_oid(1, 2, 840, 10045, 3, 1, 7),
        )
        self.pending_public_key = self.pending_signing_key.verifying_key.to_string(
            encoding="uncompressed"
        )
        public_key = der.encode_bitstring(self.pending_public_key, unused=0)
        spki = der.encode_sequence(algorithm, public_key)
        certification_request_info.append(spki)

        # Extensions
        extension_request = der.encode_sequence(
            der.encode_oid(1, 2, 840, 113549, 1, 9, 14),
            encode_set(der.encode_sequence()),
        )
        certification_request_info.append(der.encode_constructed(0, extension_request))

        certification_request_info = der.encode_sequence(*certification_request_info)
        certification_request.append(certification_request_info)

        signature_algorithm = der.encode_sequence(
            der.encode_oid(1, 2, 840, 10045, 4, 3, 2)
        )
        certification_request.append(signature_algorithm)

        # Signature
        signature = self.pending_signing_key.sign_deterministic(
            certification_request_info,
            hashfunc=hashlib.sha256,
            sigencode=ecdsa.util.sigencode_der_canonize,
        )
        certification_request.append(der.encode_bitstring(signature, unused=0))

        # Generate a new key pair.
        new_key_csr = der.encode_sequence(*certification_request)

        # Create a CSR to reply back with. Sign it with the new private key.
        elements = NOCSRElements()
        elements.csr = new_key_csr
        elements.CSRNonce = args.CSRNonce
        elements = elements.encode()
        nocsr_tbs = elements.tobytes() + session.attestation_challenge

        # class CSRResponse(tlv.Structure):
        #     NOCSRElements = tlv.OctetStringMember(0, RESP_MAX)
        #     AttestationSignature = tlv.OctetStringMember(1, 64)
        response = NodeOperationalCredentialsCluster.CSRResponse()
        response.NOCSRElements = elements
        response.AttestationSignature = self.dac_key.sign_deterministic(
            nocsr_tbs, hashfunc=hashlib.sha256, sigencode=ecdsa.util.sigencode_string
        )
        return response

    def add_trusted_root_certificate(
        self,
        session,
        args: NodeOperationalCredentialsCluster.AddTrustedRootCertificate,
    ) -> interaction_model.StatusCode:
        self.pending_root_cert = args.RootCACertificate
        return interaction_model.StatusCode.SUCCESS

    def add_noc(
        self, session, args: NodeOperationalCredentialsCluster.AddNOC
    ) -> NodeOperationalCredentialsCluster.NOCResponse:
        # Section 11.18.6.8
        noc = crypto.MatterCertificate.decode(args.NOCValue)

        response = NodeOperationalCredentialsCluster.NOCResponse()

        if noc.ec_pub_key != self.pending_public_key:
            print(noc.ec_pub_key, self.pending_public_key)
            response.StatusCode = NodeOperationalCertStatusEnum.INVALID_PUBLIC_KEY
            return response

        # Save info about the fabric.
        new_fabric_index = len(self.fabrics)
        if new_fabric_index >= self.supported_fabrics:
            response.StatusCode = NodeOperationalCertStatusEnum.TABLE_FULL
            return response

        session.local_fabric_index = new_fabric_index

        # Store the NOC.
        noc_struct = NodeOperationalCredentialsCluster.NOCStruct()
        noc_struct.NOC = args.NOCValue
        if args.ICACValue:
            noc_struct.ICAC = args.ICACValue
        self.nocs.append(noc_struct)

        # Get the root cert public key so we can create the compressed fabric id.
        root_cert = crypto.MatterCertificate.decode(self.pending_root_cert)

        # Store the fabric
        new_fabric = NodeOperationalCredentialsCluster.FabricDescriptorStruct()
        new_fabric.RootPublicKey = root_cert.ec_pub_key
        new_fabric.VendorID = args.AdminVendorId
        new_fabric.FabricID = noc.subject.matter_fabric_id
        new_fabric.NodeID = noc.subject.matter_node_id
        print(f"Adding fabric {new_fabric.FabricID} with node id {new_fabric.NodeID:x}")
        self.fabrics.append(new_fabric)

        new_group_key = GroupKeyManagementCluster.KeySetWrite()
        key_set = GroupKeySetStruct()
        key_set.GroupKeySetID = 0
        key_set.GroupKeySecurityPolicy = GroupKeySetSecurityPolicyEnum.TRUST_FIRST
        key_set.EpochKey0 = args.IPKValue
        key_set.EpochStartTime0 = 0

        new_group_key.GroupKeySet = key_set
        self.group_key_manager.key_set_write(session, new_group_key)

        self.commissioned_fabrics += 1

        self.noc_keys.append(self.pending_signing_key)
        self.encoded_noc_keys.append(
            binascii.b2a_base64(
                self.pending_signing_key.to_string(), newline=False
            ).decode("utf-8")
        )
        self._nonvolatile["pk"] = self.encoded_noc_keys

        self.trusted_root_certificates.append(self.pending_root_cert)

        self.root_certs.append(root_cert)
        fabric_id = struct.pack(">Q", noc.subject.matter_fabric_id)
        self.compressed_fabric_ids.append(
            crypto.KDF(root_cert.ec_pub_key[1:], fabric_id, b"CompressedFabric", 64)
        )
        compressed_fabric_id = self.compressed_fabric_ids[-1].hex().upper()

        node_id = struct.pack(">Q", new_fabric.NodeID).hex().upper()
        instance_name = f"{compressed_fabric_id}-{node_id}"
        self.mdns_server.advertise_service(
            "_matter",
            "_tcp",
            self.port,
            instance_name=instance_name,
            subtypes=[
                f"_I{compressed_fabric_id}._sub._matter._tcp",
            ],
        )

        response.StatusCode = NodeOperationalCertStatusEnum.OK
        return response

    def remove_fabric(
        self,
        session,
        args: NodeOperationalCredentialsCluster.RemoveFabric,
    ) -> NodeOperationalCredentialsCluster.NOCResponse:
        index = args.FabricIndex
        self.commissioned_fabrics -= 1

        self.noc_keys[index] = None
        self.root_certs[index] = None
        self.compressed_fabric_ids[index] = None
        self.fabrics[index] = None
        self.nocs[index] = None

        response = NodeOperationalCredentialsCluster.NOCResponse()
        response.StatusCode = NodeOperationalCertStatusEnum.OK
        return response


class _GroupKeyManagementCluster(GroupKeyManagementCluster):
    def __init__(self):
        super().__init__()
        self.key_sets = []
        self._encoded_key_sets = []

    def restore(self, nonvolatile):
        super().restore(nonvolatile)

        if "gks" not in nonvolatile:
            return

        self._encoded_key_sets = nonvolatile["gks"]
        self.key_sets = [
            GroupKeySetStruct.decode(binascii.a2b_base64(v)) for v in nonvolatile["gks"]
        ]

    def key_set_write(
        self, session, args: GroupKeyManagementCluster.KeySetWrite
    ) -> interaction_model.StatusCode:
        self.key_sets.append(args.GroupKeySet)
        self._encoded_key_sets.append(
            binascii.b2a_base64(args.GroupKeySet.encode(), newline=False).decode(
                "utf-8"
            )
        )
        self._nonvolatile["gks"] = self._encoded_key_sets
        return interaction_model.StatusCode.SUCCESS


class RootNode(simple_device.SimpleDevice):
    DEVICE_TYPE_ID = 0x0011
    REVISION = 2

    def __init__(
        self,
        random_source,
        mdns_server,
        port,
        vendor_id,
        product_id,
        version="",
        serial_number="1234",
    ):
        super().__init__("root")

        basic_info = BasicInformationCluster()
        basic_info.vendor_id = vendor_id
        basic_info.product_id = product_id
        basic_info.product_name = "CircuitMatter"
        basic_info.serial_number = serial_number
        basic_info.software_version_string = version
        self.servers.append(basic_info)
        access_control = AccessControlCluster()
        self.servers.append(access_control)
        group_keys = _GroupKeyManagementCluster()
        self.servers.append(group_keys)
        network_info = NetworkCommissioningCluster()
        network_info.feature_map = (
            NetworkCommissioningCluster.FeatureBitmap.WIFI_NETWORK_INTERFACE
        )

        ethernet = NetworkCommissioningCluster.NetworkInfoStruct()
        ethernet.NetworkID = "enp13s0".encode("utf-8")
        ethernet.Connected = True
        network_info.networks = [ethernet]
        network_info.scan_max_time_seconds = 10
        network_info.connect_max_time_seconds = 10
        network_info.supported_wifi_bands = [
            NetworkCommissioningCluster.WifiBandEnum.BAND_2G4
        ]
        network_info.last_network_status = (
            NetworkCommissioningCluster.NetworkCommissioningStatus.SUCCESS
        )
        network_info.last_network_id = ethernet.NetworkID
        self.servers.append(network_info)
        general_commissioning = _GeneralCommissioningCluster()
        self.servers.append(general_commissioning)
        self.noc = _NodeOperationalCredentialsCluster(
            group_keys, random_source, mdns_server, port
        )
        self.servers.append(self.noc)

        self.general_diagnostics = GeneralDiagnosticsCluster()
        self.servers.append(self.general_diagnostics)

        self.user_label = user_label.UserLabelCluster()
        self.servers.append(self.user_label)

    @property
    def fabric_count(self):
        return self.noc.commissioned_fabrics



import subprocess


class Avahi:
    def __init__(self):
        self.active_services = {}

    def advertise_service(
        self,
        service_type,
        protocol,
        port,
        txt_records={},
        subtypes=[],
        instance_name="",
    ):
        subtypes = [f"--subtype={subtype}" for subtype in subtypes]
        txt_records = [f"{key}={value}" for key, value in txt_records.items()]
        command = [
            "avahi-publish-service",
            *subtypes,
            instance_name,
            f"{service_type}.{protocol}",
            str(port),
            *txt_records,
        ]
        self.active_services[service_type + instance_name] = subprocess.Popen(command)

    def __del__(self):
        for active_service in self.active_services.values():
            active_service.kill()



import enum

from . import protocol
from . import tlv


class StatusCode(enum.IntEnum):
    SUCCESS = 0x00
    """Operation was successful."""
    FAILURE = 0x01
    """Operation was not successful."""
    INVALID_SUBSCRIPTION = 0x7D
    """Subscription ID is not active."""
    UNSUPPORTED_ACCESS = 0x7E
    """The sender of the action or command does not have authorization or access."""
    UNSUPPORTED_ENDPOINT = 0x7F
    """The endpoint indicated is unsupported on the node."""
    INVALID_ACTION = 0x80
    """The action is malformed, has missing fields, or fields with invalid values. Action not carried out."""
    UNSUPPORTED_COMMAND = 0x81
    """The indicated command ID is not supported on the cluster instance. Command not carried out."""
    INVALID_COMMAND = 0x85
    """The cluster command is malformed, has missing fields, or fields with invalid values. Command not carried out."""
    UNSUPPORTED_ATTRIBUTE = 0x86
    """The indicated attribute ID, field ID or list entry does not exist for an attribute path."""
    CONSTRAINT_ERROR = 0x87
    """Out of range error or set to a reserved value. Attribute keeps its old value. Note that an attribute value may be out of range if an attribute is related to another, e.g. with minimum and maximum attributes. See the individual attribute descriptions for specific details."""
    UNSUPPORTED_WRITE = 0x88
    """Attempt to write a read-only attribute."""
    RESOURCE_EXHAUSTED = 0x89
    """An action or operation failed due to insufficient available resources."""
    NOT_FOUND = 0x8B
    """The indicated data field or entry could not be found."""
    UNREPORTABLE_ATTRIBUTE = 0x8C
    """Reports cannot be issued for this attribute."""
    INVALID_DATA_TYPE = 0x8D
    """The data type indicated is undefined or invalid for the indicated data field. Command or action not carried out."""
    UNSUPPORTED_READ = 0x8F
    """Attempt to read a write-only attribute."""
    DATA_VERSION_MISMATCH = 0x92
    """Cluster instance data version did not match request path."""
    TIMEOUT = 0x94
    """The transaction was aborted due to time being exceeded."""
    UNSUPPORTED_NODE = 0x9B
    """The node ID indicated is not supported on the node."""
    BUSY = 0x9C
    """The receiver is busy processing another action that prevents the execution of the incoming action."""
    UNSUPPORTED_CLUSTER = 0xC3
    """The cluster indicated is not supported on the endpoint."""
    NO_UPSTREAM_SUBSCRIPTION = 0xC5
    """Used by proxies to convey to clients the lack of an upstream subscription to a source."""
    NEEDS_TIMED_INTERACTION = 0xC6
    """A Untimed Write or Untimed Invoke interaction was used for an attribute or command that requires a Timed Write or Timed Invoke."""
    UNSUPPORTED_EVENT = 0xC7
    """The indicated event ID is not supported on the cluster instance."""
    PATHS_EXHAUSTED = 0xC8
    """The receiver has insufficient resources to support the specified number of paths in the request."""
    TIMED_REQUEST_MISMATCH = 0xC9
    """A request with TimedRequest field set to TRUE was issued outside a Timed transaction or a request with TimedRequest set to FALSE was issued inside a Timed transaction."""
    FAILSAFE_REQUIRED = 0xCA
    """A request requiring a Fail-safe context was invoked without the Fail-Safe context."""
    INVALID_IN_STATE = 0xCB
    """The received request cannot be handled due to the current operational state of the device."""
    NO_COMMAND_RESPONSE = 0xCC
    """A CommandDataIB is missing a response in the InvokeResponses of an Invoke Response action."""


class AttributePathIB(tlv.List):
    """Section 10.6.2"""

    EnableTagCompression = tlv.BoolMember(0, optional=True)
    Node = tlv.IntMember(1, signed=False, octets=8, optional=True)
    Endpoint = tlv.IntMember(2, signed=False, octets=2, optional=True)
    Cluster = tlv.IntMember(3, signed=False, octets=4, optional=True)
    Attribute = tlv.IntMember(4, signed=False, octets=4, optional=True)
    ListIndex = tlv.IntMember(5, signed=False, octets=2, nullable=True, optional=True)
    WildcardPathFlags = tlv.IntMember(6, signed=False, octets=4, optional=True)


class EventPathIB(tlv.List):
    """Section 10.6.8"""

    Node = tlv.IntMember(0, signed=False, octets=8)
    Endpoint = tlv.IntMember(1, signed=False, octets=2)
    Cluster = tlv.IntMember(2, signed=False, octets=4)
    Event = tlv.IntMember(3, signed=False, octets=4)
    IsUrgent = tlv.BoolMember(4)


class EventFilterIB(tlv.Structure):
    """Section 10.6.6"""

    Node = tlv.IntMember(0, signed=False, octets=8)
    EventMinimumInterval = tlv.IntMember(1, signed=False, octets=8)


class ClusterPathIB(tlv.List):
    Node = tlv.IntMember(0, signed=False, octets=8)
    Endpoint = tlv.IntMember(1, signed=False, octets=2)
    Cluster = tlv.IntMember(2, signed=False, octets=4)


class DataVersionFilterIB(tlv.Structure):
    Path = tlv.StructMember(0, ClusterPathIB)
    DataVersion = tlv.IntMember(1, signed=False, octets=4)


class StatusIB(tlv.Structure):
    Status = tlv.EnumMember(0, StatusCode)
    ClusterStatus = tlv.IntMember(1, signed=False, octets=1, optional=True)


class AttributeDataIB(tlv.Structure):
    DataVersion = tlv.IntMember(0, signed=False, octets=4)
    Path = tlv.ListMember(1, AttributePathIB)
    Data = tlv.AnythingMember(2)


class AttributeStatusIB(tlv.Structure):
    Path = tlv.ListMember(0, AttributePathIB)
    Status = tlv.StructMember(1, StatusIB)


class AttributeReportIB(tlv.Structure):
    AttributeStatus = tlv.StructMember(0, AttributeStatusIB, optional=True)
    AttributeData = tlv.StructMember(1, AttributeDataIB, optional=True)


class InteractionModelMessage(tlv.Structure):
    PROTOCOL_ID = protocol.ProtocolId.INTERACTION_MODEL

    InteractionModelRevision = tlv.IntMember(0xFF, signed=False, octets=1, default=11)


class ChunkedMessage(InteractionModelMessage):
    """Chunked messages take multiple encodes or decodes before they are complete."""

    def encode_into(self, buffer: memoryview, offset: int = 0) -> int:
        # Leave room for MoreChunkedMessages, SupressResponse, and InteractionModelRevision.
        buffer[0] = tlv.ElementType.STRUCTURE
        offset += 1
        subbuffer = memoryview(buffer)[: -2 * 2 - 3 - 1]
        del self.MoreChunkedMessages
        for name, descriptor_class in self._members():
            if isinstance(descriptor_class, tlv.ArrayMember):
                try:
                    offset = descriptor_class.encode_into(self, subbuffer, offset)
                except tlv.ArrayEncodingError as e:
                    print("splitting", name, f"[{e.index}:] offset {offset}")
                    offset = e.offset
                    tag = descriptor_class.tag
                    self.values[tag] = self.values[tag][e.index :]
                    self.MoreChunkedMessages = True
            else:
                offset = descriptor_class.encode_into(self, buffer, offset)
        buffer[offset] = tlv.ElementType.END_OF_CONTAINER
        return offset + 1


class ReadRequestMessage(InteractionModelMessage):
    PROTOCOL_OPCODE = protocol.InteractionModelOpcode.READ_REQUEST

    AttributeRequests = tlv.ArrayMember(0, AttributePathIB)
    EventRequests = tlv.ArrayMember(1, EventPathIB)
    EventFilters = tlv.ArrayMember(2, EventFilterIB)
    FabricFiltered = tlv.BoolMember(3)
    DataVersionFilters = tlv.ArrayMember(4, DataVersionFilterIB)


class WriteRequestMessage(ChunkedMessage):
    PROTOCOL_OPCODE = protocol.InteractionModelOpcode.WRITE_REQUEST

    SuppressResponse = tlv.BoolMember(0, optional=True)
    TimedRequest = tlv.BoolMember(1)
    WriteRequests = tlv.ArrayMember(2, AttributeDataIB)
    MoreChunkedMessages = tlv.BoolMember(3, optional=True)


class WriteResponseMessage(InteractionModelMessage):
    PROTOCOL_OPCODE = protocol.InteractionModelOpcode.WRITE_RESPONSE

    WriteResponses = tlv.ArrayMember(0, AttributeStatusIB)


class EventStatusIB(tlv.Structure):
    Path = tlv.StructMember(0, EventPathIB)
    Status = tlv.StructMember(1, StatusIB)


class EventDataIB(tlv.Structure):
    Path = tlv.StructMember(0, EventPathIB)
    EventNumber = tlv.IntMember(1, signed=False, octets=8)
    PriorityLevel = tlv.IntMember(2, signed=False, octets=1)

    # Only one of the below values
    EpochTimestamp = tlv.IntMember(3, signed=False, octets=8, optional=True)
    SystemTimestamp = tlv.IntMember(4, signed=False, octets=8, optional=True)
    DeltaEpochTimestamp = tlv.IntMember(5, signed=True, octets=8, optional=True)
    DeltaSystemTimestamp = tlv.IntMember(6, signed=True, octets=8, optional=True)

    Data = tlv.AnythingMember(7)


class EventReportIB(tlv.Structure):
    EventStatus = tlv.StructMember(0, EventStatusIB)
    EventData = tlv.StructMember(1, EventDataIB)


class ReportDataMessage(ChunkedMessage):
    PROTOCOL_OPCODE = protocol.InteractionModelOpcode.REPORT_DATA

    SubscriptionId = tlv.IntMember(0, signed=False, octets=4, optional=True)
    AttributeReports = tlv.ArrayMember(1, AttributeReportIB, optional=True)
    EventReports = tlv.ArrayMember(2, EventReportIB, optional=True)
    MoreChunkedMessages = tlv.BoolMember(3, optional=True)
    SuppressResponse = tlv.BoolMember(4, optional=True)


class CommandPathIB(tlv.List):
    Endpoint = tlv.IntMember(0, signed=False, octets=2)
    Cluster = tlv.IntMember(1, signed=False, octets=4)
    Command = tlv.IntMember(2, signed=False, octets=4)


class CommandDataIB(tlv.Structure):
    CommandPath = tlv.ListMember(0, CommandPathIB)
    CommandFields = tlv.AnythingMember(1, optional=True)
    CommandRef = tlv.NumberMember(2, "H", optional=True)


class CommandStatusIB(tlv.Structure):
    CommandPath = tlv.ListMember(0, CommandPathIB)
    Status = tlv.StructMember(1, StatusIB)
    CommandRef = tlv.NumberMember(2, "H", optional=True)


class InvokeResponseIB(tlv.Structure):
    Command = tlv.StructMember(0, CommandDataIB, optional=True)
    Status = tlv.StructMember(1, CommandStatusIB, optional=True)


class InvokeRequestMessage(InteractionModelMessage):
    PROTOCOL_OPCODE = protocol.InteractionModelOpcode.INVOKE_REQUEST

    SuppressResponse = tlv.BoolMember(0)
    TimedRequest = tlv.BoolMember(1)
    InvokeRequests = tlv.ArrayMember(2, CommandDataIB)


class InvokeResponseMessage(InteractionModelMessage):
    PROTOCOL_OPCODE = protocol.InteractionModelOpcode.INVOKE_RESPONSE

    SuppressResponse = tlv.BoolMember(0)
    InvokeResponses = tlv.ArrayMember(1, InvokeResponseIB)
    MoreChunkedMessages = tlv.BoolMember(2, optional=True)


class SubscribeRequestMessage(InteractionModelMessage):
    PROTOCOL_OPCODE = protocol.InteractionModelOpcode.SUBSCRIBE_REQUEST

    KeepSubscriptions = tlv.BoolMember(0)
    MinIntervalFloor = tlv.IntMember(1, signed=False, octets=2)
    MaxIntervalCeiling = tlv.IntMember(2, signed=False, octets=2)
    AttributeRequests = tlv.ArrayMember(3, AttributePathIB, optional=True)
    EventRequests = tlv.ArrayMember(4, EventPathIB, optional=True)
    EventFilters = tlv.ArrayMember(5, EventFilterIB, optional=True)
    FabricFiltered = tlv.BoolMember(7)
    DataVersionFilters = tlv.ArrayMember(8, DataVersionFilterIB, optional=True)


class StatusResponseMessage(InteractionModelMessage):
    PROTOCOL_OPCODE = protocol.InteractionModelOpcode.STATUS_RESPONSE

    Status = tlv.EnumMember(0, StatusCode)


class SubscribeResponseMessage(InteractionModelMessage):
    PROTOCOL_OPCODE = protocol.InteractionModelOpcode.SUBSCRIBE_RESPONSE

    SubscriptionId = tlv.IntMember(0, signed=False, octets=4)
    MaxInterval = tlv.IntMember(2, signed=False, octets=2)



from circuitmatter import tlv
from circuitmatter.data_model import (
    Cluster,
    ClusterId,
    DeviceTypeId,
    EndpointNumber,
    Enum8,
    List,
    ListAttribute,
    NumberAttribute,
    Uint64,
)


class AccessControlEntryPrivilegeEnum(Enum8):
    VIEW = 1
    """Can read and observe all (except Access Control Cluster and as seen by a non-Proxy)"""
    PROXY_VIEW = 2
    """Can read and observe all (as seen by a Proxy)"""
    OPERATE = 3
    """View privileges, and can perform the primary function of this Node (except Access Control Cluster)"""
    MANAGE = 4
    """Operate privileges, and can modify persistent configuration of this Node (except Access Control Cluster)"""
    ADMINISTER = 5
    """Manage privileges, and can observe and modify the Access Control Cluster"""


class AccessControlEntryAuthModeEnum(Enum8):
    PASE = 1
    """Passcode authenticated session"""
    CASE = 2
    """Certificate authenticated session"""
    GROUP = 3
    """Group authenticated session"""


class AccessControlTargetStruct(tlv.Structure):
    Cluster = ClusterId(0)
    Endpoint = EndpointNumber(1)
    DeviceType = DeviceTypeId(2)


class AccessControlCluster(Cluster):
    CLUSTER_ID = 0x001F

    class AccessControlEntryStruct(tlv.Structure):
        Privilege = tlv.EnumMember(1, AccessControlEntryPrivilegeEnum)
        AuthMode = tlv.EnumMember(2, AccessControlEntryAuthModeEnum)
        Subjects = List(3, Uint64())
        Targets = List(4, AccessControlTargetStruct, nullable=True)

    class AccessControlExtensionStruct(tlv.Structure):
        Data = tlv.OctetStringMember(1, max_length=128)

    ACL = ListAttribute(0x0000, AccessControlEntryStruct, default=[])
    Extension = ListAttribute(0x0001, AccessControlExtensionStruct, optional=True)
    SubjectsPerAccessControlEntry = NumberAttribute(
        0x0002, signed=False, bits=16, default=4
    )
    TargetsPerAccessControlEntry = NumberAttribute(
        0x0003, signed=False, bits=16, default=3
    )
    AccessControlEntriesPerFabric = NumberAttribute(
        0x0004, signed=False, bits=16, default=4
    )



from circuitmatter.clusters.system_model import binding, descriptor, user_label


class SimpleDevice:
    def __init__(self, name):
        self.name = name
        self.servers = []
        self.descriptor = descriptor.DescriptorCluster()
        device_types = []
        for superclass in type(self).__mro__:
            if not hasattr(superclass, "DEVICE_TYPE_ID"):
                continue
            device_type = descriptor.DescriptorCluster.DeviceTypeStruct()
            device_type.DeviceType = superclass.DEVICE_TYPE_ID
            device_type.Revision = superclass.REVISION
            device_types.append(device_type)
        self.descriptor.DeviceTypeList = device_types
        self.descriptor.PartsList = []
        self.descriptor.ServerList = []
        self.descriptor.ClientList = []
        self.servers.append(self.descriptor)

        self.binding = binding.BindingCluster()
        self.servers.append(self.binding)

        self.user_label = user_label.UserLabelCluster()
        self.servers.append(self.user_label)

    def restore(self, nonvolatile):
        """Restore device state from the nonvolatile dictionary and hang onto it for any updates."""
        self.nonvolatile = nonvolatile
        for server in self.servers:
            cluster_hex = hex(server.CLUSTER_ID)
            if cluster_hex not in nonvolatile:
                nonvolatile[cluster_hex] = {}
            server.restore(nonvolatile[cluster_hex])



from circuitmatter import certificates, pase


def test_basic():
    vendor_id = 0xFFF4
    product_id = 0x1234
    discriminator = 2721
    passcode = 42430398
    assert (
        pase.compute_qr_code(vendor_id, product_id, discriminator, passcode)
        == "MNOA5D4V0163Z072Y00"
    )
    assert certificates.compute_manual_code(discriminator, passcode) == "2449-902-5895"


def test_min():
    vendor_id = 0xFFF4
    product_id = 0x1234
    discriminator = 0
    passcode = 1
    assert (
        pase.compute_qr_code(vendor_id, product_id, discriminator, passcode)
        == "MNOA55UM00ID0000000"
    )
    assert certificates.compute_manual_code(discriminator, passcode) == "0000-010-0007"


def test_max():
    vendor_id = 0xFFF4
    product_id = 0x1234
    discriminator = 0xFFF
    passcode = 99999998
    assert (
        pase.compute_qr_code(vendor_id, product_id, discriminator, passcode)
        == "MNOA5N15271DQ36B420"
    )
    assert certificates.compute_manual_code(discriminator, passcode) == "3575-986-1036"



import enum
import json
import time

from . import case
from . import crypto
from . import protocol
from . import tlv
from .exchange import Exchange
from .message import ExchangeFlags, SecurityFlags

import cryptography
from cryptography.hazmat.primitives.ciphers.aead import AESCCM
import ecdsa
import hashlib
import pathlib
import struct


# Section 4.11.2
MSG_COUNTER_WINDOW_SIZE = 32
MSG_COUNTER_SYNC_REQ_JITTER_MS = 500
MSG_COUNTER_SYNC_TIMEOUT_MS = 400


class SessionParameterStruct(tlv.Structure):
    session_idle_interval = tlv.IntMember(1, signed=False, octets=4, optional=True)
    session_active_interval = tlv.IntMember(2, signed=False, octets=4, optional=True)
    session_active_threshold = tlv.IntMember(3, signed=False, octets=2, optional=True)
    data_model_revision = tlv.IntMember(4, signed=False, octets=2)
    interaction_model_revision = tlv.IntMember(5, signed=False, octets=2)
    specification_version = tlv.IntMember(6, signed=False, octets=4)
    max_paths_per_invoke = tlv.IntMember(7, signed=False, octets=2)


class GeneralCode(enum.IntEnum):
    SUCCESS = 0
    """Operation completed successfully."""

    FAILURE = 1
    """Generic failure, additional details may be included in the protocol specific status."""

    BAD_PRECONDITION = 2
    """Operation was rejected by the system because the system is in an invalid state."""

    OUT_OF_RANGE = 3
    """A value was out of a required range"""

    BAD_REQUEST = 4
    """A request was unrecognized or malformed"""

    UNSUPPORTED = 5
    """An unrecognized or unsupported request was received"""

    UNEXPECTED = 6
    """A request was not expected at this time"""

    RESOURCE_EXHAUSTED = 7
    """Insufficient resources to process the given request"""

    BUSY = 8
    """Device is busy and cannot handle this request at this time"""

    TIMEOUT = 9
    """A timeout occurred"""

    CONTINUE = 10
    """Context-specific signal to proceed"""

    ABORTED = 11
    """Failure, may be due to a concurrency error."""

    INVALID_ARGUMENT = 12
    """An invalid/unsupported argument was provided"""

    NOT_FOUND = 13
    """Some requested entity was not found"""

    ALREADY_EXISTS = 14
    """The sender attempted to create something that already exists"""

    PERMISSION_DENIED = 15
    """The sender does not have sufficient permissions to execute the requested operations."""

    DATA_LOSS = 16
    """Unrecoverable data loss or corruption has occurred."""

    MESSAGE_TOO_LARGE = 17
    """Message size is larger than the recipient can handle."""


class SecureChannelProtocolCode(enum.IntEnum):
    SESSION_ESTABLISHMENT_SUCCESS = 0x0000
    """Indication that the last session establishment message was successfully processed."""

    NO_SHARED_TRUST_ROOTS = 0x0001
    """Failure to find a common set of shared roots."""

    INVALID_PARAMETER = 0x0002
    """Generic failure during session establishment."""

    CLOSE_SESSION = 0x0003
    """Indication that the sender will close the current session."""

    BUSY = 0x0004
    """Indication that the sender cannot currently fulfill the request."""


class StatusReport:
    PROTOCOL_ID = protocol.ProtocolId.SECURE_CHANNEL
    PROTOCOL_OPCODE = protocol.SecureProtocolOpcode.STATUS_REPORT

    def __init__(self):
        self.clear()

    def clear(self):
        self.general_code: GeneralCode = 0
        self.protocol_id = 0
        self.protocol_code = 0
        self.protocol_data = None

    def __len__(self):
        return 8 + len(self.protocol_data) if self.protocol_data else 0

    def encode_into(self, buffer, offset=0) -> int:
        struct.pack_into(
            "<HIH",
            buffer,
            offset,
            self.general_code,
            self.protocol_id,
            self.protocol_code,
        )
        offset += 8
        if self.protocol_data:
            buffer[offset : offset + len(self.protocol_data)] = self.protocol_data
            offset += len(self.protocol_data)
        return offset

    def decode(self, buffer):
        self.general_code, self.protocol_id, self.protocol_code = struct.unpack_from(
            "<HIH", buffer
        )
        self.general_code = GeneralCode(self.general_code)
        self.protocol_data = buffer[8:]
        if self.protocol_id in protocol.ProtocolId:
            self.protocol_id = protocol.ProtocolId(self.protocol_id)

        if self.protocol_id == protocol.ProtocolId.SECURE_CHANNEL:
            self.protocol_code = SecureChannelProtocolCode(self.protocol_code)

    def __str__(self):
        return f"StatusReport: General Code: {self.general_code!r}, Protocol ID: {self.protocol_id!r}, Protocol Code: {self.protocol_code!r}, Protocol Data: {self.protocol_data.hex() if self.protocol_data else None}"


class UnsecuredSessionContext:
    def __init__(
        self,
        socket,
        message_counter,
        initiator,
        ephemeral_initiator_node_id,
        node_ipaddress,
    ):
        self.socket = socket
        self.initiator = initiator
        self.ephemeral_initiator_node_id = ephemeral_initiator_node_id
        self.message_reception_state = None
        self.message_counter = message_counter
        self.node_ipaddress = node_ipaddress
        self.exchanges = {}

        self.local_node_id = 0

    def send(self, message):
        message.flags |= 1  # DSIZ = 1 for destination node
        message.destination_node_id = self.ephemeral_initiator_node_id
        if message.message_counter is None:
            message.message_counter = next(self.message_counter)
        buf = memoryview(bytearray(1280))
        nbytes = message.encode_into(buf)
        self.socket.sendto(buf[:nbytes], self.node_ipaddress)


class SecureSessionContext:
    def __init__(self, random_source, socket, local_session_id):
        self.session_type = None
        """Records whether the session was established using CASE or PASE."""
        self.session_role_initiator = False
        """Records whether the node is the session initiator or responder."""
        self.local_session_id = local_session_id
        """Individually selected by each participant in secure unicast communication during session establishment and used as a unique identifier to recover encryption keys, authenticate incoming messages and associate them to existing sessions."""
        self.peer_session_id = None
        """Assigned by the peer during session establishment"""
        self.i2r_key = None
        """Encrypts data in messages sent from the initiator of session establishment to the responder."""
        self.r2i_key = None
        """Encrypts data in messages sent from the session establishment responder to the initiator."""
        self.shared_secret = None
        """Computed during the CASE protocol execution and re-used when CASE session resumption is implemented."""
        self.local_message_counter = MessageCounter(random_source=random_source)
        """Secure Session Message Counter for outbound messages."""
        self.message_reception_state = None
        """Provides tracking for the Secure Session Message Counter of the remote"""
        self.local_fabric_index = None
        """Records the local Index for the session’s Fabric, which MAY be used to look up Fabric metadata related to the Fabric for which this session context applies."""
        self.peer_node_id = 0
        """Records the authenticated node ID of the remote peer, when available."""
        self.resumption_id = None
        """The ID used when resuming a session between the local and remote peer."""
        self.session_timestamp = None
        """A timestamp indicating the time at which the last message was sent or received. This timestamp SHALL be initialized with the time the session was created."""
        self.active_timestamp = None
        """A timestamp indicating the time at which the last message was received. This timestamp SHALL be initialized with the time the session was created."""
        self.session_idle_interval = None
        self.session_active_interval = None
        self.session_active_threshold = None
        self.exchanges = {}

        self.local_node_id = 0

        self._nonce = bytearray(crypto.AEAD_NONCE_LENGTH_BYTES)
        self.socket = socket
        self.node_ipaddress = None

    def __str__(self):
        return f"Secure Session #{self.local_session_id} with {self.peer_node_id:x}"

    @property
    def peer_active(self):
        return (time.monotonic() - self.active_timestamp) < self.session_active_interval

    def decrypt_and_verify(self, message):
        cipher = self.i2r
        if self.session_role_initiator:
            cipher = self.r2i
        try:
            # TODO: Support group messages
            struct.pack_into(
                "<BIQ",
                self._nonce,
                0,
                message.security_flags,
                message.message_counter,
                self.peer_node_id,
            )
            decrypted_payload = cipher.decrypt(
                self._nonce, bytes(message.payload), bytes(message.header)
            )
        except cryptography.exceptions.InvalidTag:
            return False

        message.decrypted = True
        message.payload = decrypted_payload
        return True

    def send(self, message):
        message.session_id = self.peer_session_id
        cipher = self.r2i
        if self.session_role_initiator:
            cipher = self.i2r

        self.session_timestamp = time.monotonic()

        message.destination_node_id = self.peer_node_id
        if message.message_counter is None:
            message.message_counter = next(self.local_message_counter)

        buf = memoryview(bytearray(1280))
        nbytes = message.encode_into(buf, cipher)
        self.socket.sendto(buf[:nbytes], self.node_ipaddress)


class MessageReceptionState:
    def __init__(self, starting_value, rollover=True, encrypted=False):
        """Implements 4.6.5.1"""
        self.message_counter = starting_value
        self.window_bitmap = (1 << MSG_COUNTER_WINDOW_SIZE) - 1
        self.mask = self.window_bitmap
        self.encrypted = encrypted
        self.rollover = rollover

    def process_counter(self, counter) -> bool:
        """Returns True if the counter number is a duplicate"""
        # Process the current window first. Behavior outside the window varies.
        if counter == self.message_counter:
            return True
        if self.message_counter <= MSG_COUNTER_WINDOW_SIZE < counter:
            # Window wraps
            bit_position = 0xFFFFFFFF - counter + self.message_counter
        else:
            bit_position = self.message_counter - counter - 1
        if 0 <= bit_position < MSG_COUNTER_WINDOW_SIZE:
            if self.window_bitmap & (1 << bit_position) != 0:
                # This is a duplicate message
                return True
            self.window_bitmap |= 1 << bit_position
            return False

        new_start = (self.message_counter + 1) & self.mask  # Inclusive
        new_end = (
            self.message_counter - MSG_COUNTER_WINDOW_SIZE
        ) & self.mask  # Exclusive
        if not self.rollover:
            new_end = (1 << MSG_COUNTER_WINDOW_SIZE) - 1
        elif self.encrypted:
            new_end = (
                self.message_counter + (1 << (MSG_COUNTER_WINDOW_SIZE - 1))
            ) & self.mask

        if new_start <= new_end:
            if not (new_start <= counter < new_end):
                return True
        else:
            if not (counter < new_end or new_start <= counter):
                return True

        # This is a new message
        shift = counter - self.message_counter
        if counter < self.message_counter:
            shift += 0x100000000
        if shift > MSG_COUNTER_WINDOW_SIZE:
            self.window_bitmap = 0
        else:
            new_bitmap = (self.window_bitmap << shift) & self.mask
            self.window_bitmap = new_bitmap
        if 1 <= shift < MSG_COUNTER_WINDOW_SIZE:
            self.window_bitmap |= 1 << (shift - 1)
        self.message_counter = counter
        return False


class MessageCounter:
    def __init__(self, starting_value=None, random_source=None):
        if starting_value is None:
            starting_value = random_source.urandom(4)
            starting_value = struct.unpack("<I", starting_value)[0]
            starting_value >>= 4
            starting_value += 1
        self.value = starting_value

    def __next__(self):
        self.value = (self.value + 1) % 0xFFFFFFFF
        return self.value


class SessionManager:
    def __init__(self, random_source, socket, node_credentials):
        persist_path = pathlib.Path("counters.json")
        if persist_path.exists():
            self.nonvolatile = json.loads(persist_path.read_text())
        else:
            self.nonvolatile = {}
            self.nonvolatile["check_in_counter"] = None
            self.nonvolatile["group_encrypted_data_message_counter"] = None
            self.nonvolatile["group_encrypted_control_message_counter"] = None
        self.unencrypted_message_counter = MessageCounter(random_source=random_source)
        self.group_encrypted_data_message_counter = MessageCounter(
            self.nonvolatile["group_encrypted_data_message_counter"],
            random_source=random_source,
        )
        self.group_encrypted_control_message_counter = MessageCounter(
            self.nonvolatile["group_encrypted_control_message_counter"],
            random_source=random_source,
        )
        self.check_in_counter = MessageCounter(
            self.nonvolatile["check_in_counter"], random_source=random_source
        )
        self.unsecured_session_context = {}
        self.secure_session_contexts = ["reserved"]
        self.socket = socket
        self.random = random_source
        self.node_credentials = node_credentials

    def _increment(self, value):
        return (value + 1) % 0xFFFFFFFF

    def get_session(self, message):
        if message.secure_session:
            if message.security_flags & SecurityFlags.GROUP:
                if message.source_node_id is None:
                    return None
                # TODO: Get MRS for source node id and message type
            else:
                session_context = self.secure_session_contexts[message.session_id]
                session_context.node_ipaddress = message.source_ipaddress
        else:
            if message.source_node_id not in self.unsecured_session_context:
                self.unsecured_session_context[message.source_node_id] = (
                    UnsecuredSessionContext(
                        self.socket,
                        self.unencrypted_message_counter,
                        initiator=False,
                        ephemeral_initiator_node_id=message.source_node_id,
                        node_ipaddress=message.source_ipaddress,
                    )
                )
            session_context = self.unsecured_session_context[message.source_node_id]
        return session_context

    def mark_duplicate(self, message):
        """Implements 4.6.7"""
        session_context = self.get_session(message)

        if session_context.message_reception_state is None:
            session_context.message_reception_state = MessageReceptionState(
                message.message_counter,
                rollover=False,
                encrypted=message.secure_session,
            )
            message.duplicate = False
            return

        message.duplicate = session_context.message_reception_state.process_counter(
            message.message_counter
        )

    def next_message_counter(self, message):
        """Implements 4.6.6"""
        if not message.secure_session:
            value = self.unencrypted_message_counter
            self.unencrypted_message_counter = self._increment(
                self.unencrypted_message_counter
            )
            return value
        elif message.security_flags & SecurityFlags.GROUP:
            if message.security_flags & SecurityFlags.C:
                value = self.group_encrypted_control_message_counter
                self.group_encrypted_control_message_counter = self._increment(
                    self.group_encrypted_control_message_counter
                )
                return value
            else:
                value = self.group_encrypted_data_message_counter
                self.group_encrypted_data_message_counter = self._increment(
                    self.group_encrypted_data_message_counter
                )
                return value
        session = self.secure_session_contexts[message.session_id]
        value = session.local_message_counter
        next_value = self._increment(value)
        session.local_message_counter = next_value
        if next_value == 0:
            # TODO expire the encryption key
            raise NotImplementedError("Expire the encryption key 4.6.6")
        return next_value

    def new_context(self):
        if None not in self.secure_session_contexts:
            self.secure_session_contexts.append(None)
        session_id = self.secure_session_contexts.index(None)

        self.secure_session_contexts[session_id] = SecureSessionContext(
            self.random, self.socket, session_id
        )
        return self.secure_session_contexts[session_id]

    def process_exchange(self, message):
        session = self.get_session(message)
        if session is None:
            return None
        # Step 1 of 4.12.5.2
        if (
            message.exchange_flags & (ExchangeFlags.R | ExchangeFlags.A)
            and not message.security_flags & SecurityFlags.C
            and message.security_flags & SecurityFlags.GROUP
        ):
            # Drop illegal combination of flags.
            return None
        if message.exchange_id not in session.exchanges:
            # Section 4.10.5.2
            initiator = message.exchange_flags & ExchangeFlags.I
            if initiator and not message.duplicate:
                session.exchanges[message.exchange_id] = Exchange(
                    session, not initiator, message.exchange_id, [message.protocol_id]
                )
                # Drop because the message isn't from an initiator.
            elif message.exchange_flags & ExchangeFlags.R:
                # Send a bare acknowledgement back.
                raise NotImplementedError("Send a bare acknowledgement back")
                return None
            else:
                # Just drop it.
                return None

        exchange = session.exchanges[message.exchange_id]
        if exchange.receive(message):
            # If we want to drop the message, then return None.
            return None

        return exchange

    def reply_to_sigma1(self, exchange, sigma1):
        if sigma1.resumptionID is None != sigma1.initiatorResumeMIC is None:
            print("Invalid resumption ID")
            error_status = StatusReport()
            error_status.general_code = GeneralCode.FAILURE
            error_status.protocol_id = protocol.ProtocolId.SECURE_CHANNEL
            error_status.protocol_code = SecureChannelProtocolCode.INVALID_PARAMETER
            return error_status
        if sigma1.resumptionID is not None:
            # Resume
            print("Ignoring resumptionID")

        matching_noc = None
        identity_protection_key = b""
        for i, fabric in enumerate(self.node_credentials.fabrics):
            root_public_key = self.node_credentials.root_certs[i].ec_pub_key
            key_set = self.node_credentials.group_key_manager.key_sets[i]
            compressed_fabric_id = self.node_credentials.compressed_fabric_ids[i]
            ipk_epoch_key = key_set.EpochKey0
            identity_protection_key = crypto.KDF(
                ipk_epoch_key,
                compressed_fabric_id,
                b"GroupKey v1.0",
                crypto.SYMMETRIC_KEY_LENGTH_BITS,
            )
            fabric_id = struct.pack("<Q", fabric.FabricID)
            node_id = struct.pack("<Q", fabric.NodeID)
            candidate_destination_id = case.compute_destination_id(
                root_public_key,
                fabric_id,
                node_id,
                sigma1.initiatorRandom,
                identity_protection_key,
            )
            if sigma1.destinationId == candidate_destination_id:
                matching_noc = i
                break

        if matching_noc is None:
            error_status = StatusReport()
            error_status.general_code = GeneralCode.FAILURE
            error_status.protocol_id = protocol.ProtocolId.SECURE_CHANNEL
            error_status.protocol_code = SecureChannelProtocolCode.NO_SHARED_TRUST_ROOTS
            return error_status

        fabric = self.node_credentials.fabrics[matching_noc]

        session_context = self.new_context()
        exchange.secure_session_context = session_context
        session_context.session_role_initiator = False
        session_context.peer_session_id = sigma1.initiatorSessionId
        session_context.local_fabric_index = matching_noc + 1
        session_context.resumption_id = self.random.urandom(16)
        session_context.local_node_id = fabric.NodeID

        ephemeral_key_pair = ecdsa.keys.SigningKey.generate(
            curve=ecdsa.NIST256p, hashfunc=hashlib.sha256, entropy=self.random.urandom
        )

        ephemeral_public_key = ephemeral_key_pair.verifying_key.to_string(
            encoding="uncompressed"
        )

        session_context.shared_secret = crypto.ECDH(
            ephemeral_key_pair, sigma1.initiatorEphPubKey
        )

        tbsdata = case.Sigma2TbsData()
        tbedata = case.Sigma2TbeData()

        tbsdata.responderNOC = self.node_credentials.nocs[matching_noc].NOC
        tbedata.responderNOC = self.node_credentials.nocs[matching_noc].NOC

        icac = self.node_credentials.nocs[matching_noc].ICAC
        if icac:
            tbsdata.responderICAC = self.node_credentials.nocs[matching_noc].ICAC
            tbedata.responderICAC = self.node_credentials.nocs[matching_noc].ICAC

        tbsdata.responderEphPubKey = ephemeral_public_key
        tbsdata.initiatorEphPubKey = sigma1.initiatorEphPubKey

        tbsdata = tbsdata.encode()

        tbedata.signature = self.node_credentials.noc_keys[
            matching_noc
        ].sign_deterministic(
            tbsdata,
            hashfunc=hashlib.sha256,
            sigencode=ecdsa.util.sigencode_string,
        )
        tbedata.resumptionID = session_context.resumption_id

        random = self.random.urandom(32)
        exchange.transcript_hash = hashlib.sha256(sigma1.encode())
        salt = (
            identity_protection_key
            + random
            + ephemeral_public_key
            + exchange.transcript_hash.digest()
        )
        s2k = crypto.KDF(
            session_context.shared_secret,
            salt,
            b"Sigma2",
            crypto.SYMMETRIC_KEY_LENGTH_BITS,
        )

        sigma2 = case.Sigma2()
        sigma2.responderRandom = random
        sigma2.responderSessionId = session_context.local_session_id
        sigma2.responderEphPubKey = ephemeral_public_key

        s2k_cipher = AESCCM(
            s2k,
            tag_length=crypto.AEAD_MIC_LENGTH_BYTES,
        )
        sigma2.encrypted2 = s2k_cipher.encrypt(
            b"NCASE_Sigma2N", bytes(tbedata.encode()), b""
        )

        exchange.transcript_hash.update(sigma2.encode())
        exchange.identity_protection_key = identity_protection_key
        exchange.s3k = crypto.KDF(
            session_context.shared_secret,
            identity_protection_key + exchange.transcript_hash.digest(),
            b"Sigma3",
            crypto.SYMMETRIC_KEY_LENGTH_BITS,
        )
        return sigma2

    def reply_to_sigma3(self, exchange, sigma3) -> SecureChannelProtocolCode:
        s3k_cipher = AESCCM(
            exchange.s3k,
            tag_length=crypto.AEAD_MIC_LENGTH_BYTES,
        )
        try:
            decrypted = s3k_cipher.decrypt(b"NCASE_Sigma3N", sigma3.encrypted3, b"")
        except cryptography.exceptions.InvalidTag:
            return SecureChannelProtocolCode.INVALID_PARAMETER
        sigma3_tbe = case.Sigma3TbeData.decode(decrypted)

        # TODO: Implement checks 4a-4d. INVALID_PARAMETER if they fail.

        # TODO: Verify NOC chain. Checks 6a-6b. INVALID_PARAMETER if they fail.

        # TODO: Verify with TBS data. Steps 8 and 9. INVALID_PARAMETER if they fail.

        secure_session_context = exchange.secure_session_context

        peer_noc = sigma3_tbe.initiatorNOC
        peer_noc = crypto.MatterCertificate.decode(peer_noc)
        secure_session_context.peer_node_id = peer_noc.subject.matter_node_id

        exchange.transcript_hash.update(sigma3.encode())

        # Generate session keys
        keys = crypto.KDF(
            secure_session_context.shared_secret,
            exchange.identity_protection_key + exchange.transcript_hash.digest(),
            b"SessionKeys",
            3 * crypto.SYMMETRIC_KEY_LENGTH_BITS,
        )
        secure_session_context.i2r_key = keys[: crypto.SYMMETRIC_KEY_LENGTH_BYTES]
        secure_session_context.i2r = AESCCM(
            secure_session_context.i2r_key,
            tag_length=crypto.AEAD_MIC_LENGTH_BYTES,
        )
        secure_session_context.r2i_key = keys[
            crypto.SYMMETRIC_KEY_LENGTH_BYTES : 2 * crypto.SYMMETRIC_KEY_LENGTH_BYTES
        ]
        secure_session_context.r2i = AESCCM(
            secure_session_context.r2i_key,
            tag_length=crypto.AEAD_MIC_LENGTH_BYTES,
        )
        secure_session_context.attestation_challenge = keys[
            2 * crypto.SYMMETRIC_KEY_LENGTH_BYTES :
        ]

        secure_session_context.session_timestamp = time.monotonic()
        return SecureChannelProtocolCode.SESSION_ESTABLISHMENT_SUCCESS
